# This file is part of Elide, frontend to Lisien, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""The big widget that shows the graph of the selected Character."""

from functools import partial
from time import monotonic

from kivy.clock import Clock, mainthread, triggered
from kivy.logger import Logger
from kivy.properties import (
	BooleanProperty,
	DictProperty,
	ListProperty,
	NumericProperty,
	ObjectProperty,
	ReferenceListProperty,
	StringProperty,
)
from kivy.uix.floatlayout import FloatLayout
from kivy.uix.image import Image
from kivy.uix.relativelayout import RelativeLayout
from kivy.uix.widget import Widget
from kivy.vector import Vector

from lisien.util import normalize_layout

from ..boardscatter import BoardScatterPlane
from ..boardview import BoardView
from ..dummy import Dummy
from ..kivygarden.texturestack import Stack, TextureStackPlane
from ..util import store_kv
from .arrow import (
	DEFAULT_ARROW_LABEL_KWARGS,
	ArrowPlane,
	GraphArrow,
	GraphArrowWidget,
	get_points_multi,
)
from .pawn import Pawn
from .spot import GraphSpot


def trigger(func):
	return triggered()(func)


class KvLayoutBack(FloatLayout):
	"""What to show behind the graph.

	By default, shows nothing.

	"""


class KvLayoutFront(FloatLayout):
	"""What to show in front of the graph.

	By default, shows nothing.

	"""

	pass


class FinalLayout(FloatLayout):
	def finalize_all(self, *args):
		for child in self.children:
			child.finalize()
		self.bind(children=self._trigger_finalize_all)

	_trigger_finalize_all = trigger(finalize_all)


def get_label_kwargs_from_portal(portal):
	return {
		"text": str(portal.get(portal.get("_label_stat", None), "")),
		**portal.get("_label_kwargs", {}),
		**DEFAULT_ARROW_LABEL_KWARGS,
	}


class GraphBoard(RelativeLayout):
	"""A graphical view onto a :class:`lisien.Character`, resembling a game
	graph.

	"""

	app = ObjectProperty()
	character = ObjectProperty()
	wallpaper_path = StringProperty()
	spot = DictProperty({})
	pawn = DictProperty({})
	arrow = DictProperty({})
	pred_arrow = DictProperty({})
	wallpaper = ObjectProperty()
	kvlayoutback = ObjectProperty()
	arrow_plane = ObjectProperty()
	stack_plane = ObjectProperty()
	kvlayoutfront = ObjectProperty()
	wids = ReferenceListProperty(
		wallpaper, kvlayoutback, arrow_plane, stack_plane, kvlayoutfront
	)
	spots_unposd = ListProperty([])
	layout_tries = NumericProperty(5)
	tracking_vel = BooleanProperty(False)
	selection_candidates = ListProperty([])
	selection = ObjectProperty(allownone=True)
	keep_selection = ObjectProperty(False)
	adding_portal = BooleanProperty(False)
	reciprocal_portal = BooleanProperty(False)
	grabbing = BooleanProperty(True)
	grabbed = ObjectProperty(None, allownone=True)
	spot_cls = ObjectProperty(Stack)
	pawn_cls = ObjectProperty(Stack)
	arrow_cls = GraphArrow
	proto_arrow_cls = ObjectProperty(GraphArrowWidget)
	_scheduled_rm_spot = DictProperty()
	_scheduled_rm_arrow = DictProperty()
	_scheduled_discard_pawn = DictProperty()
	_scheduled_add_pawn = DictProperty()

	@property
	def widkwargs(self):
		return {"size_hint": (None, None), "size": self.size, "pos": (0, 0)}

	def on_touch_down(self, touch):
		"""Check for collisions and select an appropriate entity."""
		if hasattr(self, "_lasttouch") and self._lasttouch == touch:
			return
		touch.push()
		touch.apply_transform_2d(self.to_local)
		if not self.collide_point(*touch.pos):
			touch.pop()
			return
		if self.app.selection:
			if self.app.selection.collide_point(*touch.pos):
				Logger.debug("Board: hit selection")
				if hasattr(self.app.selection, "__self__"):
					touch.grab(self.app.selection)
			else:
				if hasattr(self.app.selection, "selected"):
					self.app.selection.selected = False
				self.app.selection = None
		pawns = list(self.pawns_at(*touch.pos))
		if pawns:
			Logger.debug("Board: hit {} pawns".format(len(pawns)))
			self.selection_candidates = pawns
			if self.app.selection in self.selection_candidates:
				self.selection_candidates.remove(self.app.selection)
			touch.pop()
			return True
		spots = list(self.spots_at(*touch.pos))
		if spots:
			Logger.debug("Board: hit {} spots".format(len(spots)))
			self.selection_candidates = spots
			if self.adding_portal:
				self.origspot = self.selection_candidates.pop(0)
				self.protodest = Dummy(
					name="protodest", pos=touch.pos, size=(0, 0)
				)
				self.add_widget(self.protodest)
				self.protodest.on_touch_down(touch)
				self.protoportal = self.proto_arrow_cls(
					board=self,
					origin=self.origspot,
					destination=self.protodest,
				)
				self.add_widget(self.protoportal)
				if self.reciprocal_portal:
					self.protoportal2 = self.proto_arrow_cls(
						board=self,
						destination=self.origspot,
						origin=self.protodest,
					)
					self.add_widget(self.protoportal2)
			touch.pop()
			return True
		edges = list(self.arrow_plane.iter_collided_edges(*touch.pos))
		if edges:
			Logger.debug("Board: hit {} arrows".format(len(edges)))
			self.selection_candidates = [
				self.arrow[orig][dest] for (orig, dest) in edges
			]
			if self.app.selection in self.selection_candidates:
				self.selection_candidates.remove(self.app.selection)
			if (
				isinstance(self.app.selection, GraphArrow)
				and self.app.selection.reciprocal in self.selection_candidates
			):
				self.selection_candidates.remove(self.app.selection.reciprocal)
			touch.pop()
			return True
		touch.pop()

	def on_touch_move(self, touch):
		"""If an entity is selected, drag it."""
		if hasattr(self, "_lasttouch") and self._lasttouch == touch:
			return
		if self.app.selection in self.selection_candidates:
			self.selection_candidates.remove(self.app.selection)
		if self.app.edit_locked:
			return
		elif self.app.selection:
			if not self.selection_candidates:
				self.keep_selection = True
			ret = super().on_touch_move(touch)
			sel = self.app.selection
			if isinstance(sel, Stack):
				sel.center = touch.pos
				name = sel.proxy.name
				touch.ud["grabbed_pawn_or_spot"] = name
				for dest in self.arrow.get(name, ()):
					self.arrow[name][dest].repoint()
				for orig in self.pred_arrow.get(name, ()):
					self.arrow[orig][name].repoint()
				for thing in sel.proxy.contents():
					pawn = self.pawn[thing["name"]]
					pawn.pos = sel.right, sel.top  # real layout needed
			return ret
		elif self.selection_candidates:
			for cand in self.selection_candidates:
				if cand.collide_point(*touch.pos):
					self.app.selection = cand
					cand.selected = True
					if isinstance(cand, Widget):
						touch.grab(cand)
					ret = super().on_touch_move(touch)
					return ret
		if hasattr(self, "protodest"):
			self.protodest.pos = touch.pos
			self.protoportal._trigger_repoint()
			if hasattr(self, "protoportal2"):
				self.protoportal2._trigger_repoint()

	def portal_touch_up(self, touch):
		"""Try to create a portal between the spots the user chose."""
		try:
			# If the touch ended upon a spot, and there isn't
			# already a portal between the origin and this
			# destination, create one.
			destspot = next(self.spots_at(*touch.pos))
			orig = self.origspot.proxy
			dest = destspot.proxy
			if orig != dest and not (
				orig.name in self.character.portal
				and dest.name in self.character.portal[orig.name]
			):
				symmetrical = hasattr(self, "protoportal2") and not (
					orig.name in self.character.preportal
					and dest.name in self.character.preportal[orig.name]
				)
				port = self.character.new_portal(
					orig.name, dest.name, symmetrical=symmetrical
				)
				self.arrow_plane.add_new_portal(self.make_arrow(port))
				if orig.name not in self.arrow:
					self.arrow[orig.name] = {}
				if dest.name not in self.pred_arrow:
					self.pred_arrow[dest.name] = {}
				self.arrow[orig.name][dest.name] = self.pred_arrow[dest.name][
					orig.name
				] = GraphArrow(
					board=self,
					origin=self.spot[orig.name],
					destination=self.spot[dest.name],
				)
				if symmetrical:
					self.arrow_plane.add_new_portal(
						self.make_arrow(
							self.character.portal[dest.name][orig.name]
						)
					)
					if dest.name not in self.arrow:
						self.arrow[dest.name] = {}
					if orig.name not in self.pred_arrow:
						self.pred_arrow[orig.name] = {}
					self.arrow[dest.name][orig.name] = self.pred_arrow[
						orig.name
					][dest.name] = GraphArrow(
						board=self,
						origin=self.spot[dest.name],
						destination=self.spot[orig.name],
					)
		except StopIteration:
			pass
		self.remove_widget(self.protoportal)
		del self.protoportal
		if hasattr(self, "protoportal2"):
			self.remove_widget(self.protoportal2)
			del self.protoportal2
		del self.protodest

	def on_touch_up(self, touch):
		"""Delegate touch handling if possible, else select something."""

		def unsel_graph_arrow():
			origspot = self.app.selection.origin
			destspot = self.app.selection.destination
			insts = self.arrow_plane._instructions_map[
				origspot.name, destspot.name
			]
			fbo = self.arrow_plane._fbo
			fbo.bind()
			insts["color0"].rgba = self.arrow_plane.bg_color_unselected
			insts["color1"].rgba = self.arrow_plane.fg_color_unselected
			fbo.clear_buffer()
			fbo.release()
			self.arrow_plane.canvas.ask_update()

		if hasattr(self, "_lasttouch") and self._lasttouch == touch:
			return
		self._lasttouch = touch
		touch.push()
		touch.apply_transform_2d(self.to_local)
		if hasattr(self, "protodest"):
			Logger.debug("Board: on_touch_up making a portal")
			touch.ungrab(self)
			ret = self.portal_touch_up(touch)
			touch.pop()
			return ret
		if self.app.selection:
			sel = self.app.selection
			if isinstance(sel, Widget):
				sel.dispatch("on_touch_up", touch)
			elif (
				isinstance(sel, Stack)
				and hasattr(sel, "proxy")
				and hasattr(sel.proxy, "name")
				and "grabbed_pawn_or_spot" in touch.ud
				and sel.proxy.name == touch.ud["grabbed_pawn_or_spot"]
				and not self.app.edit_locked
			):
				if hasattr(sel.proxy, "location"):
					for candidate in self.stack_plane.iter_collided_keys(
						*touch.pos
					):
						if candidate in self.spot:
							newloc = self.character.place[candidate]
							sel.proxy.location = newloc
							newspot = self.spot[candidate]
							sel.pos = newspot.right, newspot.top
							return
					oldloc = sel.proxy.location
					oldspot = self.spot[oldloc.name]
					sel.pos = oldspot.right, oldspot.top
					return
				else:
					prox = sel.proxy
					prox["_x"] = sel.x / self.width
					prox["_y"] = sel.y / self.height
		for candidate in self.selection_candidates:
			if candidate.collide_point(*touch.pos):
				if isinstance(candidate, GraphArrow):
					portal = self.character.portal[candidate.origin.name][
						candidate.destination.name
					]
					insts = self.arrow_plane._instructions_map[
						portal["origin"], portal["destination"]
					]
					fbo = self.arrow_plane._fbo
					fbo.bind()
					insts["color0"].rgba = self.arrow_plane.bg_color_selected
					insts["color1"].rgba = self.arrow_plane.fg_color_selected
					fbo.clear_buffer()
					fbo.release()
					self.arrow_plane.canvas.ask_update()
				if hasattr(candidate, "selected"):
					candidate.selected = True
				if (
					hasattr(self.app.selection, "selected")
					and self.app.selection != candidate
				):
					self.app.selection.selected = False
				if isinstance(self.app.selection, GraphArrow):
					unsel_graph_arrow()
				self.app.selection = candidate
				self.keep_selection = True
				break
		if not self.keep_selection:
			Logger.debug("Board: deselecting " + repr(self.app.selection))
			if hasattr(self.app.selection, "selected"):
				self.app.selection.selected = False
			if isinstance(self.app.selection, GraphArrow):
				unsel_graph_arrow()
			self.app.selection = None
		self.keep_selection = False
		touch.ungrab(self)
		touch.pop()
		return

	def _pull_size(self, *args):
		if self.wallpaper.texture is None:
			Clock.schedule_once(self._pull_size, 0.001)
			return
		self.size = self.wallpaper.size = self.wallpaper.texture.size

	def _pull_image(self, *args):
		self.wallpaper.source = self.wallpaper_path
		self._pull_size()

	def on_parent(self, *args):
		"""Create some subwidgets and trigger the first update."""
		if not self.parent or hasattr(self, "_parented"):
			return
		if not self.wallpaper_path:
			Logger.debug("Board: waiting for wallpaper_path")
			Clock.schedule_once(self.on_parent, 0)
			return
		self._parented = True
		self.wallpaper = Image(source=self.wallpaper_path)
		self.bind(wallpaper_path=self._pull_image)
		self._pull_size()
		self.kvlayoutback = KvLayoutBack(**self.widkwargs)
		self.arrow_plane = ArrowPlane(**self.widkwargs)
		self.stack_plane = TextureStackPlane(**self.widkwargs)
		self.kvlayoutfront = KvLayoutFront(**self.widkwargs)
		for wid in self.wids:
			self.add_widget(wid)
			wid.pos = 0, 0
			wid.size = self.size
			if wid is not self.wallpaper:
				self.bind(size=wid.setter("size"))
		self.update()

	def on_character(self, *args):
		if self.character is None or self.character.engine.closed:
			return
		if self.parent is None:
			Clock.schedule_once(self.on_character, 0)
			return

		wallpaper_path = self.character.stat.setdefault(
			"wallpaper", "parchmentBasic.png"
		)
		self.engine = getattr(self.character, "engine", None)
		if (
			"_control" not in self.character.stat
			or "wallpaper" not in self.character.stat["_control"]
		):
			control = self.character.stat.get("_control", {})
			control["wallpaper"] = "textinput"
		self.wallpaper_path = wallpaper_path
		self.trigger_update()

	def update_from_stat(self, sender, *, k, v):
		if k == "wallpaper":
			self.wallpaper_path = v

	def _trigger_pull_wallpaper(self, *args, **kwargs):
		if kwargs["key"] != "wallpaper":
			return
		if hasattr(self, "_scheduled_pull_wallpaper"):
			Clock.unschedule(self._scheduled_pull_wallpaper)
		self._scheduled_pull_wallpaper = Clock.schedule_once(
			self.pull_wallpaper, 0
		)

	@trigger
	def kv_updated(self, *args):
		self.unbind(wallpaper_path=self.kvlayoutback.setter("wallpaper_path"))
		for wid in self.wids:
			self.remove_widget(wid)
		self.kvlayoutback = KvLayoutBack(
			pos=(0, 0), wallpaper_path=self.wallpaper_path
		)
		self.bind(wallpaper_path=self.kvlayoutback.setter("wallpaper_path"))
		self.kvlayoutfront = KvLayoutFront(**self.widkwargs)
		self.size = self.kvlayoutback.size
		self.kvlayoutback.bind(size=self.setter("size"))
		for wid in self.wids:
			self.add_widget(wid)

	def make_pawn(self, thing):
		"""Make a :class:`Pawn` to represent a :class:`Thing`, store it, and
		return a dict suitable for `StackPlane.add_datum`

		"""
		if thing["name"] in self.pawn:
			raise KeyError("Already have a Pawn for this Thing")
		r = self.pawn_cls(board=self, proxy=thing)
		self.pawn[thing["name"]] = r
		locspot = self.spot[thing["location"]]
		if "_image_paths" in thing:
			texs = list(thing["_image_paths"])
		else:
			texs = list(Pawn.default_image_paths)
		width = height = 0.0
		for tex in texs:
			wide, high = Image(source=tex).texture_size
			if wide > width:
				width = wide
			if high > height:
				height = high
		return {  # need to lay out multiple pawns per spot properly
			"x": int(locspot.right),
			"y": int(locspot.top),
			"width": width,
			"height": height,
			"name": thing["name"],
			"textures": texs,
		}

	def make_spot(self, place):
		"""Make a :class:`Spot` to represent a :class:`Place`, store it, and
		return a dict suitable for `StackPlane.add_datum`

		"""
		if place["name"] in self.spot:
			raise KeyError("Already have a Spot for this Place")
		self.spot[place["name"]] = self.spot_cls(board=self, proxy=place)
		if "_image_paths" in place:
			texs = list(place["_image_paths"])
		else:
			texs = list(GraphSpot.default_image_paths)
		width = height = 0.0
		for tex in texs:
			wide, high = Image(source=tex).texture_size
			if wide > width:
				width = wide
			if high > height:
				height = high
		return {
			"x": place.get("_x", 0.5),
			"y": place.get("_y", 0.5),
			"width": width,
			"height": height,
			"name": place["name"],
			"textures": texs,
		}

	def make_arrow(self, portal):
		if (
			portal["origin"] not in self.spot
			or portal["destination"] not in self.spot
		):
			raise ValueError(
				"An :class:`Arrow` should only be made after "
				"the :class:`Spot`s it connects"
			)
		if (
			portal["origin"] in self.arrow
			and portal["destination"] in self.arrow[portal["origin"]]
		):
			raise KeyError("Already have an Arrow for this Portal")
		return self._core_make_arrow(
			portal,
			self.spot[portal["origin"]],
			self.spot[portal["destination"]],
		)

	def _core_make_arrow(self, portal, origspot, destspot, points=None):
		r = {
			"board": self,
			"portal": portal,
			"origspot": origspot,
			"destspot": destspot,
			"label_kwargs": get_label_kwargs_from_portal(portal),
		}
		if points is not None:
			r["points"] = points
		return r

	def rm_arrows_to_and_from(self, name):
		if name in self.arrow.keys():
			for dest in list(self.arrow[name].keys()):
				self.rm_arrow(name, dest)
		if name in self.pred_arrow.keys():
			for orig in list(self.pred_arrow[name].keys()):
				self.rm_arrow(orig, name)

	def rm_pawn(self, name, *args):
		"""Remove the :class:`Pawn` by the given name."""
		if name not in self.pawn:
			raise KeyError("No Pawn named {}".format(name))
		# Currently there's no way to connect Pawns with Arrows but I
		# think there will be, so, insurance
		self.rm_arrows_to_and_from(name)
		pwn = self.pawn.pop(name)
		if pwn in self.selection_candidates:
			self.selection_candidates.remove(pwn)
		self.stack_plane.remove(name)

	def _trigger_rm_pawn(self, name):
		Clock.schedule_once(partial(self.rm_pawn, name), 0)

	def rm_spot(self, name, *args):
		"""Remove the :class:`Spot` by the given name."""
		if name not in self.spot:
			raise KeyError("No Spot named {}".format(name))
		spot = self.spot.pop(name)
		if spot in self.selection_candidates:
			self.selection_candidates.remove(spot)
		pawns_here = []
		for thing in spot.proxy.contents():
			pawns_here.append(self.pawn[thing.name])
		self.rm_arrows_to_and_from(name)
		self.stack_plane.remove(name)
		for pawn in pawns_here:
			self.rm_pawn(pawn.name)
		if name in self._scheduled_rm_spot:
			del self._scheduled_rm_spot[name]

	def _trigger_rm_spot(self, name):
		part = partial(self.rm_spot, name)
		if name in self._scheduled_rm_spot:
			Clock.unschedule(self._scheduled_rm_spot[name])
		self._scheduled_rm_spot[name] = Clock.schedule_once(part, 0)

	def rm_arrow(self, orig, dest, *args):
		"""Remove the :class:`Arrow` that goes from ``orig`` to ``dest``."""
		if orig not in self.arrow or dest not in self.arrow[orig]:
			raise KeyError("No Arrow from {} to {}".format(orig, dest))
		arr = self.arrow[orig].pop(dest)
		if arr in self.selection_candidates:
			self.selection_candidates.remove(arr)
		self.arrow_plane.remove_edge(orig, dest)
		if (orig, dest) in self._scheduled_rm_arrow:
			del self._scheduled_rm_arrow[orig, dest]

	def _trigger_rm_arrow(self, orig, dest):
		part = partial(self.rm_arrow, orig, dest)
		if (orig, dest) in self._scheduled_rm_arrow:
			Clock.unschedule(self._scheduled_rm_arrow[orig, dest])
		self._scheduled_rm_arrow[orig, dest] = Clock.schedule_once(part, 0)

	def graph_layout(self, graph):
		from networkx.drawing.layout import spring_layout

		return normalize_layout(spring_layout(graph))

	def discard_pawn(self, thingn, *args):
		if thingn in self.pawn:
			self.rm_pawn(thingn)
		if thingn in self._scheduled_discard_pawn:
			del self._scheduled_discard_pawn[thingn]

	def _trigger_discard_pawn(self, thing):
		part = partial(self.discard_pawn, thing)
		if thing in self._scheduled_discard_pawn:
			Clock.unschedule(self._scheduled_discard_pawn[thing])
		self._scheduled_discard_pawn[thing] = Clock.schedule_once(part, 0)

	def _remove_absent_pawns(self, *args):
		for pawn_name in list(self.pawn.keys()):
			if pawn_name not in self.character.thing:
				self.rm_pawn(pawn_name)

	def discard_spot(self, placen, *args):
		if placen in self.spot:
			self.rm_spot(placen)

	def _trigger_discard_spot(self, place):
		Clock.schedule_once(partial(self.discard_spot, place), 0)

	def _remove_absent_spots(self, *args):
		for spot_name in list(self.spot.keys()):
			if spot_name not in self.character.place:
				self.rm_spot(spot_name)

	def discard_arrow(self, orign, destn, *args):
		if orign in self.arrow and destn in self.arrow[orign]:
			self.rm_arrow(orign, destn)

	def _trigger_discard_arrow(self, orig, dest):
		Clock.schedule_once(partial(self.discard_arrow, orig, dest), 0)

	def _remove_absent_arrows(self, *args):
		for arrow_origin in list(self.arrow.keys()):
			for arrow_destination in list(self.arrow[arrow_origin].keys()):
				if (
					arrow_origin not in self.character.portal
					or arrow_destination
					not in self.character.portal[arrow_origin]
				):
					self.rm_arrow(arrow_origin, arrow_destination)

	def add_spot(self, placen, *args):
		if placen in self.character.place and placen not in self.spot:
			spotten = self.make_spot(self.character.place[placen])
			self.stack_plane.add_datum(spotten)
			self.spot[placen].pos = (
				spotten["x"] * self.width,
				spotten["y"] * self.height,
			)

	def _trigger_add_spot(self, placen):
		Clock.schedule_once(partial(self.add_spot, placen), 0)

	def _add_new_spots(self, *args):
		start_ts = monotonic()
		places2add = []
		spots_unposd = []
		nodes_patch = {}
		placemap = self.character.place
		spotmap = self.spot
		default_image_paths = GraphSpot.default_image_paths
		for place_name, place in placemap.items():
			if place_name not in spotmap:
				place = placemap[place_name]
				places2add.append(place)
		make_spot = self.make_spot
		spots_posd = []
		stack_idx = self.stack_plane._stack_index
		for place in places2add:
			spot = make_spot(place)
			if "_x" not in place or "_y" not in place:
				spots_unposd.append(spot)
			elif spot["name"] not in stack_idx:
				spots_posd.append(spot)
		if spots_unposd:
			try:
				nodes_patch_2 = self.grid_layout(spots_unposd)
			except (TypeError, ValueError):
				nodes_patch_2 = self.nx_layout(spots_unposd)
			for k, v in nodes_patch_2.items():
				if k in nodes_patch:
					nodes_patch[k].update(v)
				else:
					nodes_patch[k] = v
		if nodes_patch:
			self.engine.handle(
				command="update_nodes",
				char=self.character.name,
				patch=nodes_patch,
			)
		if spots_posd:
			self.stack_plane.unbind_uid(
				"data", self.stack_plane._redraw_bind_data_uid
			)
			self.stack_plane.data.extend(spots_posd)
			self.stack_plane.redraw()
			self.stack_plane._redraw_bind_data_uid = self.stack_plane.fbind(
				"data", self.stack_plane._trigger_redraw
			)

	def add_arrow(self, orign, destn, *args):
		if not (
			orign in self.character.portal
			and destn in self.character.portal[orign]
		):
			raise ValueError("No portal for arrow {}->{}".format(orign, destn))
		portal = self.character.portal[orign][destn]
		if not (orign in self.arrow and destn in self.arrow[orign]):
			self.arrow_plane.add_new_portal(self.make_arrow(portal))
			the_arrow = GraphArrow(
				board=self,
				origin=self.spot[orign],
				destination=self.spot[destn],
			)
			if orign not in self.arrow:
				self.arrow[orign] = {}
			self.arrow[orign][destn] = the_arrow
			if destn not in self.pred_arrow:
				self.pred_arrow[destn] = {}
			self.pred_arrow[destn][orign] = the_arrow

	def _add_new_arrows(self, *args):
		portmap = self.character.portal
		arrowmap = self.arrow
		pred_arrowmap = self.pred_arrow
		spotmap = self.spot
		append_to_arrow_plane = self.arrow_plane.data.append
		core_make_arrow = self._core_make_arrow
		todo = []
		for arrow_orig, arrow_dests in portmap.items():
			for arrow_dest, portal in arrow_dests.items():
				if (
					arrow_orig not in arrowmap
					or arrow_dest not in arrowmap[arrow_orig]
				):
					todo.append(
						(portal, spotmap[arrow_orig], spotmap[arrow_dest])
					)
					the_arr = GraphArrow(
						board=self,
						origin=spotmap[arrow_orig],
						destination=spotmap[arrow_dest],
					)
					if arrow_orig not in arrowmap:
						arrowmap[arrow_orig] = {}
					if arrow_dest not in arrowmap[arrow_orig]:
						arrowmap[arrow_orig][arrow_dest] = the_arr
					if arrow_dest not in pred_arrowmap:
						pred_arrowmap[arrow_dest] = {}
					if arrow_orig not in pred_arrowmap[arrow_dest]:
						pred_arrowmap[arrow_dest][arrow_orig] = the_arr
		points = get_points_multi(
			(origspot, destspot, 10) for (portal, origspot, destspot) in todo
		)
		for portal, origspot, destspot in todo:
			append_to_arrow_plane(
				core_make_arrow(
					portal,
					origspot,
					destspot,
					points[origspot.name, destspot.name],
				)
			)

	def update_arrow_labels(self, *args):
		portmap = self.character.portal
		arrow_plane = self.arrow_plane
		for datum in arrow_plane.data:
			portal = portmap[datum["origspot"].name][datum["destspot"].name]
			new_kwargs = get_label_kwargs_from_portal(portal)
			if new_kwargs != datum["label_kwargs"]:
				arrow_plane.update_portal(
					datum["origspot"].name, datum["destspot"].name, new_kwargs
				)

	def add_pawn(self, thingn, *args):
		if thingn not in self.character.thing:
			raise KeyError(f"No Thing for pawn: {thingn}")
		if thingn in self.pawn:
			raise KeyError(f"Already have pawn for Thing: {thingn}")
		pwn = self.make_pawn(self.character.thing[thingn])
		stacp = self.stack_plane
		stacp.add_datum(pwn)
		self.pawn[thingn].pos = pwn["x"], pwn["y"]

		if thingn in self._scheduled_add_pawn:
			del self._scheduled_add_pawn[thingn]

	def _trigger_add_pawn(self, thingn):
		part = partial(self.add_pawn, thingn)
		if thingn in self._scheduled_add_pawn:
			Clock.unschedule(self._scheduled_add_pawn[thingn])
		self._scheduled_add_pawn[thingn] = Clock.schedule_once(part, 0)

	def _add_new_pawns(self, *args):
		nodes_patch = {}
		things2add = []
		pawns_added = []
		pawnmap = self.pawn
		for thing_name, thing in self.character.thing.items():
			if thing_name not in pawnmap:
				things2add.append(thing)
		make_pawn = self.make_pawn
		for thing in things2add:
			pwn = make_pawn(thing)
			if "_image_paths" not in thing:
				nodes_patch[thing["name"]] = {
					"_image_paths": list(
						pwn.get("textures", Pawn.default_image_paths)
					)
				}
			pawns_added.append(pwn)
		if nodes_patch:
			self.character.node.patch(nodes_patch)
		self.stack_plane.unbind_uid(
			"data", self.stack_plane._redraw_bind_data_uid
		)
		self.stack_plane.data.extend(pawns_added)
		self.stack_plane.redraw()
		self.stack_plane._redraw_bind_data_uid = self.stack_plane.fbind(
			"data", self.stack_plane._trigger_redraw
		)

	def update(self, *args):
		"""Force an update to match the current state of my character.

		This polls every element of the character, and therefore
		causes me to sync with the lisien core for a long time. Avoid
		when possible.

		"""
		if not hasattr(self, "engine") or getattr(
			self.engine, "closed", False
		):
			Logger.warning(
				"Board: tried to update without a connection to a lisien core"
			)
			return
		if not self.stack_plane or not self.arrow_plane:
			self.trigger_update()
			return
		# remove widgets that don't represent anything anymore
		Logger.debug("GraphBoard: updating")
		start_ts = monotonic()
		self._disconnect_proxy_objects()
		self._remove_absent_pawns()
		self._remove_absent_spots()
		self._remove_absent_arrows()
		# add widgets to represent new stuff
		self._add_new_spots()
		self._update_spot_display()
		if self.arrow_cls:
			self._add_new_arrows()
			self._update_arrow_display()
		self._add_new_pawns()
		self._update_pawn_display()
		self._connect_proxy_objects()
		Logger.debug(
			f"GraphBoard: updated, took {monotonic() - start_ts:,.2f} seconds"
		)

	trigger_update = trigger(update)

	def _disconnect_proxy_objects(self):
		char = self.character
		char.stat.disconnect(self.update_from_character_stat)
		char.node.disconnect(self.update_from_character_node)
		char.portal.disconnect(self.update_from_character_edge)

	def _connect_proxy_objects(self):
		char = self.character
		char.stat.connect(self.update_from_character_stat)
		char.node.connect(self.update_from_character_node)
		char.portal.connect(self.update_from_character_edge)

	def update_from_character_stat(self, character, key, value):
		pass

	@mainthread
	def update_from_character_node(self, node, key, value):
		if hasattr(node, "location"):
			if not node:
				self.rm_pawn(node.name)
			elif node.name not in self.pawn:
				self.add_pawn(node.name)
			elif key == "location":
				loc = self.spot[value]
				thing = self.pawn[node.name]
				thing.pos = loc.right, loc.top
			elif key == "_image_paths":
				self.pawn[node.name].paths = value
		elif node is self.character.node or node is self.character.place:
			if value and key not in self.spot:
				self.add_spot(key)
			elif not value and key in self.spot:
				self.rm_spot(key)
		elif node is self.character.thing:
			if value and key not in self.pawn:
				self.add_pawn(key)
			elif not value and key not in self.pawn:
				self.rm_pawn(key)
		else:
			if not node:
				self.rm_spot(node.name)
			elif node.name not in self.spot:
				self.add_spot(node.name)
			elif key == "_image_paths":
				self.spot[node.name].paths = value

	def _update_spot_display(self):
		"""Change spot graphics to match the state of their place"""

	def _update_pawn_display(self):
		"""Change pawn graphics to match the state of their thing"""

	def update_from_character_edge(self, edge, key, value):
		if edge:
			if not self.arrow_plane.have_arrow(
				edge._origin, edge._destination
			):
				label_kwargs = DEFAULT_ARROW_LABEL_KWARGS.copy()
				if "_label_stat" in edge:
					label_kwargs["text"] = str(
						edge.get(edge["label_stat"], "")
					)
				self.arrow_plane.add_new_portal(
					{
						"origspot": self.spot[edge.origin.name],
						"destspot": self.spot[edge.destination.name],
						"label_kwargs": label_kwargs,
					}
				)
			if key == edge.get("_label_stat"):
				self.arrow_plane.update_portal_label(
					edge._origin, edge._destination, str(value)
				)
		else:
			self.arrow_plane.remove_edge(edge._origin, edge._destination)

	def _update_arrow_display(self):
		"""Change arrow graphics to match the state of their portal"""

	def _apply_node_layout(self, l, spot, *args):
		if self.width == 1 or self.height == 1:
			Clock.schedule_once(
				partial(self._apply_node_layout, l, spot), 0.01
			)
			return
		if not isinstance(spot, dict):
			spot = {spt["name"]: spt for spt in spot}
		node_upd = {}
		newspots = []
		for name, (x, y) in l.items():
			assert 0 <= x <= 0.99, "{} has invalid x: {}".format(name, x)
			assert 0 <= y <= 0.99, "{} has invalid y: {}".format(name, y)
			assert self.stack_plane.width == self.width
			assert self.stack_plane.height == self.height
			node_upd[name] = {"_x": x, "_y": y}
			spot[name]["x"] = x
			spot[name]["y"] = y
			newspots.append(spot[name])
		if newspots:
			self.stack_plane.unbind_uid(
				"data", self.stack_plane._redraw_bind_data_uid
			)
			self.stack_plane.data.extend(newspots)
			self.stack_plane.redraw()
			self.stack_plane._redraw_bind_data_uid = self.stack_plane.fbind(
				"data", self.stack_plane._trigger_redraw
			)
		self.spots_unposd = []
		return node_upd

	def grid_layout(self, spots, *args):
		return self._apply_node_layout(
			normalize_layout({spot["name"]: spot["name"] for spot in spots}),
			spots,
		)

	def nx_layout(self, spots, *args):
		spots_only = self.character.facade()
		for thing in list(spots_only.thing.keys()):
			del spots_only.thing[thing]
		for place in spots_only.place.keys() - set(
			spot["name"] for spot in spots
		):
			del spots_only.place[place]
		return self._apply_node_layout(self.graph_layout(spots_only), spots)

	def arrows(self):
		"""Iterate over all my arrows."""
		for o in self.arrow.values():
			for arro in o.values():
				yield arro

	def pawns_at(self, x, y):
		"""Iterate over pawns that collide the given point."""
		for name in self.pawn.keys() & set(
			self.stack_plane.iter_collided_keys(x, y)
		):
			yield self.pawn[name]

	def spots_at(self, x, y):
		"""Iterate over spots that collide the given point."""
		for name in self.spot.keys() & set(
			self.stack_plane.iter_collided_keys(x, y)
		):
			yield self.spot[name]

	def arrows_at(self, x, y):
		"""Iterate over arrows that collide the given point."""
		for orig, dest in self.arrow_plane.iter_collided_edges(x, y):
			yield self.arrow[orig][dest]


class GraphBoardScatterPlane(BoardScatterPlane):
	selection_candidates = ListProperty([])
	selection = ObjectProperty(allownone=True)
	keep_selection = BooleanProperty(False)
	board = ObjectProperty()
	adding_portal = BooleanProperty(False)
	reciprocal_portal = BooleanProperty()

	def spot_from_dummy(self, dummy):
		"""Make a real place and its spot from a dummy spot.

		Create a new :class:`graph.Spot` instance, along with the
		underlying :class:`lisien.Place` instance, and give it the name,
		position, and imagery of the provided dummy.

		"""
		(x, y) = self.to_local(*dummy.pos_up)
		x /= self.board.width
		y /= self.board.height
		self.board.character.new_place(
			dummy.name, _x=x, _y=y, _image_paths=list(dummy.paths)
		)
		dummy.num += 1

	def pawn_from_dummy(self, dummy):
		"""Make a real thing and its pawn from a dummy pawn.

		Create a new :class:`graph.Pawn` instance, along with the
		underlying :class:`lisien.Thing` instance, and give it the name,
		location, and imagery of the provided dummy.

		"""
		candidates = []
		dummy_center = self.to_local(*dummy.center)
		dummy.pos = self.to_local(*dummy.pos)
		for key in self.board.stack_plane.iter_collided_keys(*dummy_center):
			if key in self.board.spot:
				candidates.append(self.board.spot[key])
		if not candidates:
			return
		whereat = candidates.pop()
		if candidates:
			dist = Vector(*whereat.center).distance(dummy_center)
			while candidates:
				thereat = candidates.pop()
				thereto = Vector(*thereat.center).distance(dummy_center)
				if thereto < dist:
					whereat, dist = thereat, thereto
		self.board.character.new_thing(
			dummy.name, whereat.proxy.name, _image_paths=list(dummy.paths)
		)
		dummy.num += 1

	def on_board(self, *args):
		if hasattr(self, "_oldboard"):
			self.unbind(
				adding_portal=self._oldboard.setter("adding_portal"),
				reciprocal_portal=self._oldboard.setter("reciprocal_portal"),
			)
		self.clear_widgets()
		self.add_widget(self.board)
		self.board.adding_portal = self.adding_portal
		self.board.reciprocal_portal = self.reciprocal_portal
		self.bind(
			adding_portal=self.board.setter("adding_portal"),
			reciprocal_portal=self.board.setter("reciprocal_portal"),
		)
		self._oldboard = self.board


class GraphBoardView(BoardView):
	adding_portal = BooleanProperty(False)
	reciprocal_portal = BooleanProperty(True)
	engine = ObjectProperty()
	character_name = StringProperty()

	def on_character_name(self, *args):
		if (
			not self.engine
			or not self.character_name
			or self.character_name not in self.engine.character
		):
			Clock.schedule_once(self.on_character_name, 0)
			return
		character = self.engine.character[self.character_name]
		self.board = GraphBoard(character=character)


store_kv(
	__name__,
	"""
<GraphBoard>:
	app: app
	size_hint: None, None
<GraphBoardView>:
	plane: boardplane
	GraphBoardScatterPlane:
		id: boardplane
		board: root.board
		adding_portal: root.adding_portal
		reciprocal_portal: root.reciprocal_portal
		scale_min: root.scale_min
		scale_max: root.scale_max
		pos: root.pos
		size: root.size
		size_hint: None, None
""",
)



# This file is part of Elide, frontend to Lisien, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from collections import OrderedDict
from inspect import signature
from operator import attrgetter

from kivy.app import App
from kivy.clock import Clock, triggered
from kivy.logger import Logger
from kivy.properties import ObjectProperty, StringProperty
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.button import Button
from kivy.uix.floatlayout import FloatLayout
from kivy.uix.label import Label
from kivy.uix.modalview import ModalView
from kivy.uix.recycleview import RecycleView
from kivy.uix.recycleview.views import RecycleDataViewBehavior
from kivy.uix.screenmanager import Screen
from kivy.uix.tabbedpanel import TabbedPanel, TabbedPanelItem
from kivy.uix.togglebutton import ToggleButton
from kivy.uix.widget import Widget

from .card import Card, DeckBuilderScrollBar, DeckBuilderView
from .stores import FuncEditor
from .util import logwrap, store_kv


def trigger(func):
	return triggered()(func)


dbg = Logger.debug


# How do these get instantiated?
class RuleButton(ToggleButton, RecycleDataViewBehavior):
	"""A button to select a rule to edit"""

	rulesview = ObjectProperty()
	ruleslist = ObjectProperty()
	rule = ObjectProperty()

	@logwrap(section="RuleButton")
	def on_state(self, *args):
		"""If I'm pressed, unpress all other buttons in the ruleslist"""
		# This really ought to be done with the selection behavior
		if self.state == "down":
			self.rulesview.rule = self.rule
			for button in self.ruleslist.children[0].children:
				if button != self:
					button.state = "normal"


class RulesList(RecycleView):
	"""A list of rules you might want to edit

	Presented as buttons, which you can click to select one rule at a time.

	"""

	rulebook = ObjectProperty()
	rulesview = ObjectProperty()

	@logwrap(section="RulesList")
	def on_rulebook(self, *_):
		"""Make sure to update when the rulebook changes"""
		if self.rulebook is None:
			return
		self.rulebook.connect(self._trigger_redata, weak=False)
		self.redata()

	@logwrap(section="RulesList")
	def redata(self, *_):
		"""Make my data represent what's in my rulebook right now"""
		if self.rulesview is None:
			Clock.schedule_once(self.redata, 0)
			return
		data = [
			{
				"rulesview": self.rulesview,
				"rule": rule,
				"index": i,
				"ruleslist": self,
			}
			for i, rule in enumerate(self.rulebook)
		]
		self.data = data

	@logwrap(section="RulesList")
	def _trigger_redata(self, *_, **__):
		if hasattr(self, "_scheduled_redata"):
			Clock.unschedule(self._scheduled_redata)
		self._scheduled_redata = Clock.schedule_once(self.redata, 0)


class RulesView(Widget):
	"""The view to edit a rule

	Presents three tabs, one each for trigger, prereq, and action. Each has a
	deckbuilder in it with a column of used functions and a column of unused
	actions.

	"""

	rulebook = ObjectProperty()
	entity = ObjectProperty()
	rule = ObjectProperty(allownone=True)

	@property
	def engine(self):
		return App.get_running_app().engine

	@logwrap(section="RulesView")
	def on_rule(self, *args):
		"""Make sure to update when the rule changes"""
		if self.rule is None:
			return
		self.rule.connect(self._listen_to_rule)

	@logwrap(section="RulesView")
	def _listen_to_rule(self, rule, **kwargs):
		if rule is not self.rule:
			rule.disconnect(self._listen_to_rule)
			return
		if "triggers" in kwargs:
			self.pull_triggers()
		if "prereqs" in kwargs:
			self.pull_prereqs()
		if "actions" in kwargs:
			self.pull_actions()

	def __init__(self, **kwargs):
		super().__init__(**kwargs)
		self.finalize()

	@logwrap(section="RulesView")
	def finalize(self, *_):
		"""Add my tabs"""
		assert not getattr(self, "_finalized", False), "Already finalized"
		if not self.canvas:
			Clock.schedule_once(self.finalize, 0)
			return

		deck_builder_kwargs = {
			"pos_hint": {"x": 0, "y": 0},
			"starting_pos_hint": {"x": 0.05, "top": 0.95},
			"card_size_hint": (0.3, 0.4),
			"card_hint_step": (0, -0.1),
			"deck_x_hint_step": 0.4,
		}

		self._tabs = TabbedPanel(
			size=self.size, pos=self.pos, do_default_tab=False
		)
		self.bind(size=self._tabs.setter("size"), pos=self._tabs.setter("pos"))
		self.add_widget(self._tabs)

		for functyp in "trigger", "prereq", "action":
			tab = TabbedPanelItem(text=functyp.capitalize())
			setattr(self, "_{}_tab".format(functyp), tab)
			self._tabs.add_widget(getattr(self, "_{}_tab".format(functyp)))
			builder = DeckBuilderView(**deck_builder_kwargs)
			setattr(self, "_{}_builder".format(functyp), builder)
			setattr(
				self,
				f"_trigger_push_{functyp}s_uid",
				builder.fbind(
					"decks", getattr(self, f"_trigger_push_{functyp}s")
				),
			)
			scroll_left = DeckBuilderScrollBar(
				size_hint_x=0.01,
				pos_hint={"x": 0, "y": 0},
				deckbuilder=builder,
				deckidx=0,
				scroll_min=0,
			)
			setattr(self, "_scroll_left_" + functyp, scroll_left)
			scroll_right = DeckBuilderScrollBar(
				size_hint_x=0.01,
				pos_hint={"right": 1, "y": 0},
				deckbuilder=builder,
				deckidx=1,
				scroll_min=0,
			)
			setattr(self, "_scroll_right_" + functyp, scroll_right)
			layout = FloatLayout()
			setattr(self, "_{}_layout".format(functyp), layout)
			tab.add_widget(layout)
			layout.add_widget(builder)
			layout.add_widget(scroll_left)
			layout.add_widget(scroll_right)
			layout.add_widget(
				Label(
					text="Used",
					pos_hint={"center_x": 0.1, "center_y": 0.98},
					size_hint=(None, None),
				)
			)
			layout.add_widget(
				Label(
					text="Unused",
					pos_hint={"center_x": 0.5, "center_y": 0.98},
					size_hint=(None, None),
				)
			)
			self.bind(rule=getattr(self, "_trigger_pull_{}s".format(functyp)))
		self._finalized = True

	@logwrap(section="RulesView")
	def _edit_something(self, card: Card):
		what_store = card.ud["type"]
		what_function = card.headline_text
		Logger.info(f"EditButton: {what_store}.{what_function}")
		if not hasattr(self, "rule_func_editor_modal"):
			self.rule_func_editor_modal = ModalView()
			self.rule_func_editor = FuncEditor(
				size_hint_y=0.9, deletable=False
			)

			def save_and_dismiss_modal(*args):
				changed = self.rule_func_editor.save()
				Logger.info(
					f"EditButton: {what_store}.{what_function} {'not ' if not changed else ''}changed"
				)
				if changed:
					card.text = self.rule_func_editor.source
				self.rule_func_editor_modal.dismiss()

			self.rule_func_editor_save_button = Button(
				text="Save", on_press=save_and_dismiss_modal
			)
			self.rule_func_editor_cancel_button = Button(
				text="Cancel", on_press=self.rule_func_editor_modal.dismiss
			)
			self.rule_func_editor_layout = BoxLayout(orientation="vertical")
			self.rule_func_editor_buttons_layout = BoxLayout(
				orientation="horizontal", size_hint_y=0.1
			)
			self.rule_func_editor_buttons_layout.add_widget(
				self.rule_func_editor_cancel_button
			)
			self.rule_func_editor_buttons_layout.add_widget(
				self.rule_func_editor_save_button
			)
			self.rule_func_editor_layout.add_widget(self.rule_func_editor)
			self.rule_func_editor_layout.add_widget(
				self.rule_func_editor_buttons_layout
			)
			self.rule_func_editor_modal.add_widget(
				self.rule_func_editor_layout
			)
		# get the source code of the function to edit
		store = self.rule_func_editor.store = getattr(self.engine, what_store)
		self.rule_func_editor.source = store.get_source(what_function)
		self.rule_func_editor.name_wid.hint_text = what_function
		self.rule_func_editor.name_wid.disabled = True
		# show the modal
		self.rule_func_editor_modal.open()

	@logwrap(section="RulesView")
	def get_functions_cards(self, what, allfuncs):
		"""Return a pair of lists of Card widgets for used and unused functions.

		:param what: a string: 'trigger', 'prereq', or 'action'
		:param allfuncs: a sequence of functions' (name, sourcecode, signature)

		"""
		if not self.rule:
			return [], []
		rulefuncnames = list(
			map(attrgetter("name"), getattr(self.rule, what + "s"))
		)
		unused = [
			Card(
				ud={"type": what, "funcname": name, "signature": sig},
				headline_text=name,
				show_art=False,
				midline_text=what.capitalize(),
				text=source,
				editable=True,
				edit_func=self._edit_something,
			)
			for (name, source, sig) in allfuncs
			if name not in rulefuncnames
		]
		used = [
			Card(
				ud={
					"type": what,
					"funcname": name,
				},
				headline_text=name,
				show_art=False,
				midline_text=what.capitalize(),
				text=str(getattr(getattr(self.engine, what), name)),
				editable=True,
				edit_func=self._edit_something,
			)
			for name in rulefuncnames
		]
		return used, unused

	@logwrap(section="RulesView")
	def set_functions(self, what, allfuncs):
		"""Set the cards in the ``what`` builder to ``allfuncs``

		:param what: a string, 'trigger', 'prereq', or 'action'
		:param allfuncs: a sequence of triples of (name, sourcecode, signature)
		                as taken by my ``get_function_cards`` method.

		"""
		setattr(
			getattr(self, "_{}_builder".format(what)),
			"decks",
			self.get_functions_cards(what, allfuncs),
		)

	@logwrap(section="RulesView")
	def _pull_functions(self, what, truth=True):
		it = map(self.inspect_func, getattr(self.engine, what)._cache.items())
		if not truth:
			it = filter(lambda x: x[0] != "truth", it)
		return self.get_functions_cards(what, list(it))

	@logwrap(section="RulesView")
	def pull_triggers(self, *args):
		"""Refresh the cards in the trigger builder"""
		self._trigger_builder.unbind_uid(
			"decks", self._trigger_push_triggers_uid
		)
		self._trigger_builder.decks = self._pull_functions("trigger")
		self._trigger_push_triggers_uid = self._trigger_builder.fbind(
			"decks", self._trigger_push_triggers
		)

	_trigger_pull_triggers = trigger(pull_triggers)

	@logwrap(section="RulesView")
	def pull_prereqs(self, *args):
		"""Refresh the cards in the prereq builder"""
		self._prereq_builder.unbind_uid(
			"decks", self._trigger_push_prereqs_uid
		)
		self._prereq_builder.decks = self._pull_functions(
			"prereq", truth=False
		)
		self._trigger_push_prereqs_uid = self._prereq_builder.fbind(
			"decks", self._trigger_push_prereqs
		)

	_trigger_pull_prereqs = trigger(pull_prereqs)

	@logwrap(section="RulesView")
	def pull_actions(self, *args):
		"""Refresh the cards in the action builder"""
		self._action_builder.unbind_uid(
			"decks", self._trigger_push_actions_uid
		)
		self._action_builder.decks = self._pull_functions(
			"action", truth=False
		)
		self._trigger_push_actions_uid = self._action_builder.fbind(
			"decks", self._trigger_push_actions
		)

	_trigger_pull_actions = trigger(pull_actions)

	@logwrap(section="RulesView")
	def inspect_func(self, namesrc):
		"""Take a function's (name, sourcecode) and return a triple of (name, sourcecode, signature)"""
		(name, src) = namesrc
		glbls = {}
		lcls = {}
		exec(src, glbls, lcls)
		assert name in lcls
		func = lcls[name]
		return name, src, signature(func)

	@logwrap(section="RulesView")
	def update_builders(self, *args):
		for attrn in "_trigger_builder", "_prereq_builder", "_action_builder":
			if not hasattr(self, attrn):
				dbg("RulesView: no {}".format(attrn))
				Clock.schedule_once(self.update_builders, 0)
				return
		self._trigger_builder.clear_widgets()
		self._prereq_builder.clear_widgets()
		self._action_builder.clear_widgets()
		if self.rule is None:
			dbg("RulesView: no rule")
			return
		if hasattr(self, "_list"):
			self._list.redata()
		self.pull_triggers()
		self.pull_prereqs()
		self.pull_actions()

	_trigger_update_builders = trigger(update_builders)

	@logwrap(section="RulesView")
	def _upd_unused(self, what):
		"""Make sure to have exactly one copy of every valid function in the
		"unused" pile on the right.

		Doesn't read from the database.

		:param what: a string, 'trigger', 'prereq', or 'action'

		"""
		builder = getattr(self, "_{}_builder".format(what))
		updtrig = getattr(self, "_trigger_upd_unused_{}s".format(what))
		builder.unbind(decks=updtrig)
		funcs = OrderedDict()
		cards = list(self._action_builder.decks[1])
		cards.reverse()
		for card in cards:
			funcs[card.ud["funcname"]] = card
		for card in self._action_builder.decks[0]:
			if card.ud["funcname"] not in funcs:
				funcs[card.ud["funcname"]] = card.copy()
		unused = list(funcs.values())
		unused.reverse()
		builder.decks[1] = unused
		builder.bind(decks=updtrig)

	@logwrap(section="RulesView")
	def upd_unused_actions(self, *_):
		self._upd_unused("action")

	_trigger_upd_unused_actions = trigger(upd_unused_actions)

	@logwrap(section="RulesView")
	def upd_unused_triggers(self, *_):
		self._upd_unused("trigger")

	_trigger_upd_unused_triggers = trigger(upd_unused_triggers)

	@logwrap(section="RulesView")
	def upd_unused_prereqs(self, *_):
		self._upd_unused("prereq")

	_trigger_upd_unused_prereqs = trigger(upd_unused_prereqs)

	@logwrap(section="RulesView")
	def _push_funcs(self, what):
		if not self.rule:
			Logger.debug(
				"RulesView: not pushing {} for lack of rule".format(what)
			)
			return
		funcs = [
			getattr(getattr(self.engine, what), card.ud["funcname"])
			for card in getattr(self, "_{}_builder".format(what)).decks[0]
		]
		funlist = getattr(self.rule, what + "s")
		if funlist != funcs:
			setattr(self.rule, what + "s", funcs)

	@logwrap(section="RulesView")
	def push_actions(self, *_):
		self._push_funcs("action")

	_trigger_push_actions = trigger(push_actions)

	@logwrap(section="RulesView")
	def push_prereqs(self, *_):
		self._push_funcs("prereq")

	_trigger_push_prereqs = trigger(push_prereqs)

	@logwrap(section="RulesView")
	def push_triggers(self, att, *_):
		self._push_funcs("trigger")

	_trigger_push_triggers = trigger(push_triggers)


class RulesBox(BoxLayout):
	"""A BoxLayout containing a RulesList and a RulesView

	As well as an input for a new rule name; a button to add a new rule by that
	name; and a close button.

	Currently has no way to rename rules (2018-08-15)

	"""

	rulebook = ObjectProperty()
	rulebook_name = StringProperty()
	entity = ObjectProperty()
	entity_name = StringProperty()
	new_rule_name = StringProperty()
	toggle = ObjectProperty()
	ruleslist = ObjectProperty()
	rulesview = ObjectProperty()

	@property
	def engine(self):
		return App.get_running_app().engine

	@logwrap(section="RulesBox")
	def on_ruleslist(self, *_):
		if not self.ruleslist.children:
			Clock.schedule_once(self.on_ruleslist, 0)
			return
		self.ruleslist.children[0].bind(children=self._upd_ruleslist_selection)

	@logwrap(section="RulesBox")
	def new_rule(self, *_):
		if self.new_rule_name in self.engine.rule:
			# TODO: feedback to say you already have such a rule
			return
		self._new_rule_name = self.new_rule_name
		new_rule = self.engine.rule.new_empty(self.new_rule_name)
		assert new_rule is not None
		self.rulebook.append(new_rule)
		self.ruleslist.redata()
		self.ids.rulename.text = ""

	@logwrap(section="RulesBox")
	def _upd_ruleslist_selection(self, *_):
		if not hasattr(self, "_new_rule_name"):
			return
		for child in self.ruleslist.children[0].children:
			if child.text == self._new_rule_name:
				child.state = "down"
			else:
				child.state = "normal"


class RulesScreen(Screen):
	"""Screen containing a RulesBox for one rulebook"""

	entity = ObjectProperty()
	rulebook = ObjectProperty()
	toggle = ObjectProperty()
	rulesview = ObjectProperty()

	@property
	def engine(self):
		return App.get_running_app().engine

	@logwrap(section="RulesScreen")
	def new_rule(self, *_):
		self.children[0].new_rule()


class CharacterRulesScreen(Screen):
	"""Screen with TabbedPanel for all the character-rulebooks"""

	character = ObjectProperty()
	toggle = ObjectProperty()

	@logwrap(section="CharacterRulesScreen")
	def _get_rulebook(self, rb):
		return {
			"character": self.character.rulebook,
			"unit": self.character.unit.rulebook,
			"character_thing": self.character.thing.rulebook,
			"character_place": self.character.place.rulebook,
			"character_portal": self.character.portal.rulebook,
		}[rb]

	@logwrap(section="CharacterRulesScreen")
	def finalize(self, *args):
		if hasattr(self, "_finalized"):
			return
		if not (self.toggle and self.character):
			Clock.schedule_once(self.finalize, 0)
			return
		self._tabs = TabbedPanel(do_default_tab=False)
		for rb, txt in (
			("character", "character"),
			("unit", "unit"),
			("character_thing", "thing"),
			("character_place", "place"),
			("character_portal", "portal"),
		):
			tab = TabbedPanelItem(text=txt)
			setattr(self, "_{}_tab".format(rb), tab)
			box = RulesBox(
				rulebook=self._get_rulebook(rb),
				entity=self.character,
				toggle=self.toggle,
			)
			setattr(self, "_{}_box".format(rb), box)
			tab.add_widget(box)
			self._tabs.add_widget(tab)
		self.add_widget(self._tabs)
		self._finalized = True

	@logwrap(section="CharacterRulesScreen")
	def on_character(self, *_):
		if not hasattr(self, "_finalized"):
			self.finalize()
			return
		for rb in (
			"character",
			"unit",
			"character_thing",
			"character_place",
			"character_portal",
		):
			tab = getattr(self, "_{}_tab".format(rb))
			tab.content.entity = self.character
			tab.content.rulebook = self._get_rulebook(rb)
		# Currently there's no way to assign a new rulebook to an entity
		# in elide, so I don't need to account for that, but what if the
		# rulebook changes as a result of some code running in the lisien core?
		# 2018-08-13


store_kv(
	__name__,
	"""
<RuleButton>:
    text: self.rule.name if self.rule else ''
<RulesList>:
    viewclass: 'RuleButton'
    SelectableRecycleBoxLayout:
        default_size: None, dp(56)
        default_size_hint: 1, None
        height: self.minimum_height
        size_hint_y: None
        orientation: 'vertical'
<RulesBox>:
    new_rule_name: rulename.text
    ruleslist: ruleslist
    rulesview: rulesview
    rulebook_name: str(self.rulebook.name) if self.rulebook is not None else ''
    entity_name: str(self.entity.name) if self.entity is not None else ''
    orientation: 'vertical'
    Label:
        text: root.entity_name + '    -    ' + root.rulebook_name
        size_hint_y: 0.05
    BoxLayout:
        orientation: 'horizontal'
        RulesList:
            id: ruleslist
            rulebook: root.rulebook
            entity: root.entity
            rulesview: rulesview
            size_hint_x: 0.2
        RulesView:
            id: rulesview
            rulebook: root.rulebook
            entity: root.entity
            size_hint_x: 0.8
    BoxLayout:
        orientation: 'horizontal'
        size_hint_y: 0.07
        TextInput:
            id: rulename
            hint_text: 'New rule name'
            write_tab: False
            multiline: False
        Button:
            text: '+'
            on_release: root.new_rule()
        Button:
            id: closebut
            text: 'Close'
            on_release: root.toggle()
<RulesScreen>:
    name: 'rules'
    rulesview: box.rulesview
    RulesBox:
        id: box
        rulebook: root.rulebook
        entity: root.entity
        toggle: root.toggle
<CharacterRulesScreen>:
    name: 'charrules'
""",
)



# This file is part of Elide, frontend to Lisien, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from functools import partial

from kivy.properties import NumericProperty, ObjectProperty
from kivy.uix.stencilview import StencilView

from .util import logwrap

wraplog_boardview = partial(logwrap, section="BoardView")


class BoardView(StencilView):
	board = ObjectProperty()
	plane = ObjectProperty()
	scale_min = NumericProperty(allownone=True)
	scale_max = NumericProperty(allownone=True)

	@wraplog_boardview
	def spot_from_dummy(self, dummy):
		self.plane.spot_from_dummy(dummy)

	@wraplog_boardview
	def pawn_from_dummy(self, dummy):
		self.plane.pawn_from_dummy(dummy)

	@wraplog_boardview
	def on_touch_down(self, touch):
		if not self.collide_point(*touch.pos):
			return
		return super().on_touch_down(touch)



# This file is part of Elide, frontend to Lisien, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from kivy.app import App
from kivy.clock import Clock
from kivy.properties import (
	DictProperty,
	NumericProperty,
	ObjectProperty,
	StringProperty,
)
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.button import Button
from kivy.uix.dropdown import DropDown
from kivy.uix.screenmanager import Screen
from kivy.uix.textinput import TextInput

from .statlist import BaseStatListView
from .util import logwrap, store_kv


class FloatInput(TextInput):
	@logwrap(section="FloatInput")
	def insert_text(self, s, from_undo=False):
		return super().insert_text(
			"".join(c for c in s if c in "0123456789."), from_undo
		)


class ControlTypePicker(Button):
	app = ObjectProperty()
	key = ObjectProperty()
	mainbutton = ObjectProperty()
	dropdown = ObjectProperty()
	set_control = ObjectProperty()

	def __init__(self, **kwargs):
		super().__init__(**kwargs)
		self.build()

	@logwrap(section="ControlTypePicker")
	def set_value(self, k, v):
		if v is None:
			del self.app.selected_proxy[k]
		else:
			self.app.selected_proxy[k] = v

	@logwrap(section="ControlTypePicker")
	def build(self, *_):
		if None in (self.key, self.set_control):
			Clock.schedule_once(self.build, 0)
			return
		self.mainbutton = None
		self.dropdown = None
		self.dropdown = DropDown()
		self.dropdown.bind(
			on_select=lambda instance, x: self.set_control(self.key, x)
		)
		readoutbut = Button(
			text="readout",
			size_hint_y=None,
			height=self.height,
			background_color=(0.7, 0.7, 0.7, 1),
		)
		readoutbut.bind(
			on_release=lambda instance: self.dropdown.select("readout")
		)
		self.dropdown.add_widget(readoutbut)
		textinbut = Button(
			text="textinput",
			size_hint_y=None,
			height=self.height,
			background_color=(0.7, 0.7, 0.7, 1),
		)
		textinbut.bind(
			on_release=lambda instance: self.dropdown.select("textinput")
		)
		self.dropdown.add_widget(textinbut)
		togbut = Button(
			text="togglebutton",
			size_hint_y=None,
			height=self.height,
			background_color=(0.7, 0.7, 0.7, 1),
		)
		togbut.bind(
			on_release=lambda instance: self.dropdown.select("togglebutton")
		)
		self.dropdown.add_widget(togbut)
		sliderbut = Button(
			text="slider",
			size_hint_y=None,
			height=self.height,
			background_color=(0.7, 0.7, 0.7, 1),
		)
		sliderbut.bind(
			on_release=lambda instance: self.dropdown.select("slider")
		)
		self.dropdown.add_widget(sliderbut)
		self.bind(on_release=self.dropdown.open)


class ConfigListItemToggleButton(BoxLayout):
	true_text = StringProperty("0")
	false_text = StringProperty("1")

	@logwrap(section="ConfigListItemToggleButton")
	def set_true_text(self, *_):
		self.parent.set_config(
			self.parent.key, "true_text", self.ids.truetext.text
		)
		self.true_text = self.ids.truetext.text

	@logwrap(section="ConfigListItemToggleButton")
	def set_false_text(self, *_):
		self.parent.set_config(
			self.parent.key, "false_text", self.ids.falsetext.text
		)


class ConfigListItemSlider(BoxLayout):
	min = NumericProperty(0.0)
	max = NumericProperty(1.0)

	@logwrap(section="ConfigListItemSlider")
	def set_min(self, *_):
		minn = float(self.ids.minimum.text)
		try:
			self.parent.set_config(self.parent.key, "min", minn)
			self.min = minn
		except ValueError:
			self.ids.minimum.text = ""

	@logwrap(section="ConfigListItemSlider")
	def set_max(self, *_):
		maxx = float(self.ids.maximum.text)
		try:
			self.parent.set_config(self.parent.key, "max", maxx)
			self.max = maxx
			self.ids.maximum.hint_text = str(maxx)
		except ValueError:
			self.ids.maximum.text = ""


class ConfigListItemCustomizer(BoxLayout):
	key = ObjectProperty()
	control = StringProperty()
	config = DictProperty()
	set_config = ObjectProperty()

	@logwrap(section="ConfigListItemCustomizer")
	def on_control(self, *_):
		self.clear_widgets()
		if self.control == "togglebutton":
			if (
				"true_text" not in self.config
				or "false_text" not in self.config
			):
				Clock.schedule_once(self.on_control, 0)
				return
			wid = self._toggle = ConfigListItemToggleButton(
				true_text=self.config["true_text"],
				false_text=self.config["false_text"],
			)
			self.add_widget(wid)
		elif self.control == "slider":
			if "min" not in self.config or "max" not in self.config:
				Clock.schedule_once(self.on_control, 0)
				return
			wid = self._slider = ConfigListItemSlider(
				min=self.config["min"], max=self.config["max"]
			)
			self.add_widget(wid)

	@logwrap(section="ConfigListItemCustomizer")
	def on_config(self, *_):
		if hasattr(self, "_toggle"):
			if "true_text" in self.config:
				self._toggle.true_text = self.config["true_text"]
			if "false_text" in self.config:
				self._toggle.false_text = self.config["false_text"]
		if hasattr(self, "_slider"):
			if "min" in self.config:
				self._slider.min = self.config["min"]
			if "max" in self.config:
				self._slider.max = self.config["max"]


class ConfigListItem(BoxLayout):
	key = ObjectProperty()
	config = DictProperty()
	set_control = ObjectProperty()
	set_config = ObjectProperty()
	deleter = ObjectProperty()


class StatListViewConfigurator(BaseStatListView):
	statlist = ObjectProperty()
	_key_cfg_setters = DictProperty()
	_val_text_setters = DictProperty()
	_control_wids = DictProperty()

	@logwrap(section="ConfigListItemCustomizer")
	def set_control(self, key, value):
		config = self.proxy.get("_config", {})
		if value == "slider":
			if "min" not in config:
				self.set_config(key, "min", 0.0)
			if "max" not in config:
				self.set_config(key, "max", 1.0)
		elif value == "togglebutton":
			if "true_text" not in config:
				self.set_config(key, "true_text", "1")
			if "false_text" not in config:
				self.set_config(key, "false_text", "0")
		self.set_config(key, "control", value)

	@logwrap(section="ConfigListItemCustomizer")
	def munge(self, k, v):
		# makes ConfigListItem
		ret = super().munge(k, v)
		ret["deleter"] = self.del_key
		ret["set_control"] = self.set_control
		ret["set_config"] = self.set_config
		return ret


class StatScreen(Screen):
	statlist = ObjectProperty()
	statcfg = ObjectProperty()
	toggle = ObjectProperty()
	proxy = ObjectProperty()

	@property
	def engine(self):
		return App.get_running_app().engine

	@logwrap(section="ConfigListItemCustomizer")
	def new_stat(self):
		"""Look at the key and value that the user has entered into the stat
		configurator, and set them on the currently selected
		entity.

		"""
		key = self.ids.newstatkey.text
		value = self.ids.newstatval.text
		if not (key and value):
			# TODO implement some feedback to the effect that
			# you need to enter things
			return
		try:
			self.proxy[key] = self.engine.unpack(value)
		except (TypeError, ValueError):
			self.proxy[key] = value
		self.ids.newstatkey.text = ""
		self.ids.newstatval.text = ""


store_kv(
	__name__,
	"""
<ConfigListItemCustomization>:
	pos_hint: {'x': 0, 'y': 0}
<ConfigListItemToggleButton>:
	Label:
		text: 'True text:'
	TextInput:
		id: truetext
		hint_text: root.true_text
		on_text_validate: root.set_true_text()
	Label:
		text: 'False text:'
	TextInput:
		id: falsetext
		hint_text: root.false_text
		on_text_validate: root.set_false_text()
<ConfigListItemSlider>:
	Label:
		text: 'Minimum:'
	TextInput:
		id: minimum
		hint_text: str(root.min)
		multiline: False
		on_text_validate: root.set_min()
	Label:
		text: 'Maximum:'
	TextInput:
		id: maximum
		hint_text: str(root.max)
		multiline: False
		on_text_validate: root.set_max()
<ConfigListItem>:
	height: 30
	Button:
		size_hint_x: 0.4 / 3
		text: 'del'
		on_release: root.deleter(root.key)
	Label:
		size_hint_x: 0.4 / 3
		text: str(root.key)
	ControlTypePicker:
		size_hint_x: 0.4 / 3
		key: root.key
		set_control: root.set_control
		text: root.config['control'] if 'control' in root.config else 'readout'
	ConfigListItemCustomizer:
		size_hint_x: 0.6
		control: root.config['control'] if 'control' in root.config else 'readout'
		config: root.config
		key: root.key
		set_config: root.set_config
<StatScreen>:
	name: 'statcfg'
	statcfg: cfg
	BoxLayout:
		orientation: 'vertical'
		StatListViewConfigurator:
			viewclass: 'ConfigListItem'
			id: cfg
			app: app
			proxy: root.proxy
			statlist: root.statlist
			size_hint_y: 0.95
			RecycleBoxLayout:
				default_size: None, dp(56)
				default_size_hint: 1, None
				size_hint_y: None
				height: self.minimum_height
				orientation: 'vertical'
		BoxLayout:
			orientation: 'horizontal'
			size_hint_y: 0.05
			TextInput:
				id: newstatkey
				multiline: False
				write_tab: False
				hint_text: 'New stat'
			TextInput:
				id: newstatval
				multiline: False
				write_tab: False
				hint_text: 'Value'
			Button:
				id: newstatbut
				text: '+'
				on_release: root.new_stat()
			Button:
				id: closer
				text: 'Close'
				on_release: root.toggle()
""",
)



# This file is part of Elide, frontend to Lisien, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from functools import partial

from kivy.clock import Clock, triggered
from kivy.properties import ObjectProperty

from elide.kivygarden.texturestack import ImageStack

from .util import logwrap


def trigger(func):
	return triggered()(func)


class ImageStackProxy(ImageStack):
	proxy = ObjectProperty()
	name = ObjectProperty()

	@logwrap(section="ImageStackProxy")
	def finalize(self, initial=True):
		if getattr(self, "_finalized", False):
			return
		if self.proxy is None or not hasattr(self.proxy, "name"):
			Clock.schedule_once(self.finalize, 0)
			return
		if initial:
			self.name = self.proxy.name
			if "_image_paths" in self.proxy:
				try:
					self.paths = self.proxy["_image_paths"]
				except Exception as ex:
					if not ex.args[0].startswith("Unable to load image type"):
						raise ex
					self.paths = self.default_image_paths
			else:
				self.paths = self.proxy.setdefault(
					"_image_paths", self.default_image_paths
				)
			self.finalize_children(initial)
		self._paths_binding = self.fbind(
			"paths", self._trigger_push_image_paths
		)
		self._offxs_binding = self.fbind("offxs", self._trigger_push_offxs)
		self._offys_binding = self.fbind("offys", self._trigger_push_offys)
		self._finalized = True

	@logwrap(section="ImageStackProxy")
	def finalize_children(self, initial=True, *_):
		for child in self.children:
			if not getattr(child, "_finalized", False):
				child.finalize(initial=initial)

	@logwrap(section="ImageStackProxy")
	def unfinalize(self):
		self.unbind_uid("paths", self._paths_binding)
		self.unbind_uid("offxs", self._offxs_binding)
		self.unbind_uid("offys", self._offys_binding)
		self._finalized = False

	@logwrap(section="ImageStackProxy")
	def pull_from_proxy(self, *_):
		initial = not hasattr(self, "_finalized")
		self.unfinalize()
		for key, att in [
			("_image_paths", "paths"),
			("_offxs", "offxs"),
			("_offys", "offys"),
		]:
			if key in self.proxy and self.proxy[key] != getattr(self, att):
				setattr(self, att, self.proxy[key])
		self.finalize(initial)

	@logwrap(section="ImageStackProxy")
	def _trigger_pull_from_proxy(self, *args, **kwargs):
		if hasattr(self, "_scheduled_pull_from_proxy"):
			Clock.unschedule(self._scheduled_pull_from_proxy)
		self._scheduled_pull_from_proxy = Clock.schedule_once(
			self.pull_from_proxy, 0
		)

	@trigger
	@logwrap(section="ImageStackProxy")
	def _trigger_push_image_paths(self, *_):
		self.proxy["_image_paths"] = list(self.paths)

	@trigger
	@logwrap(section="ImageStackProxy")
	def _trigger_push_offxs(self, *_):
		self.proxy["_offxs"] = list(self.offxs)

	@trigger
	@logwrap(section="ImageStackProxy")
	def _trigger_push_offys(self, *_):
		self.proxy["_offys"] = list(self.offys)

	@trigger
	@logwrap(section="ImageStackProxy")
	def _trigger_push_stackhs(self, *_):
		self.proxy["_stackhs"] = list(self.stackhs)

	@trigger
	@logwrap(section="ImageStackProxy")
	def restack(self, *_):
		childs = sorted(
			list(self.children), key=lambda child: child.priority, reverse=True
		)
		self.clear_widgets()
		for child in childs:
			self.add_widget(child)
		self.do_layout()



# This file is part of Elide, frontend to Lisien, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Object to configure, start, and stop elide."""

import json
import os
import shutil
from functools import partial
from threading import Thread

from lisien.exc import OutOfTimelineError

if "KIVY_NO_ARGS" not in os.environ:
	os.environ["KIVY_NO_ARGS"] = "1"

from kivy.app import App
from kivy.clock import Clock, triggered
from kivy.logger import Logger
from kivy.properties import (
	AliasProperty,
	BooleanProperty,
	NumericProperty,
	ObjectProperty,
	StringProperty,
)
from kivy.resources import resource_add_path, resource_find
from kivy.uix.screenmanager import NoTransition, Screen, ScreenManager

import elide
import elide.charsview
import elide.dialog
import elide.menu
import elide.rulesview
import elide.screen
import elide.spritebuilder
import elide.statcfg
import elide.stores
import elide.timestream
from elide.graph.arrow import GraphArrow
from elide.graph.board import GraphBoard
from elide.grid.board import GridBoard
from elide.util import load_kv, logwrap
from lisien.proxy import (
	CharacterProxy,
	CharStatProxy,
	EngineProcessManager,
	PlaceProxy,
	ThingProxy,
)

resource_add_path(elide.__path__[0] + "/assets")
resource_add_path(elide.__path__[0] + "/assets/rltiles")
resource_add_path(elide.__path__[0] + "/assets/kenney1bit")


def trigger(func):
	return triggered()(func)


class ElideApp(App):
	"""Extensible lisien Development Environment."""

	title = "elide"

	branch = StringProperty("trunk")
	turn = NumericProperty(0)
	tick = NumericProperty(0)
	character = ObjectProperty()
	selection = ObjectProperty(None, allownone=True)
	selected_proxy = ObjectProperty()
	selected_proxy_name = StringProperty("")
	statcfg = ObjectProperty()
	edit_locked = BooleanProperty(False)
	simulate_button_down = BooleanProperty(False)
	prefix = StringProperty()
	games_dir = StringProperty("games")
	logs_dir = StringProperty(None, allownone=True)
	game_name = StringProperty("game0")
	use_thread = BooleanProperty(False)
	connect_string = StringProperty(None, allownone=True)
	workers = NumericProperty(None, allownone=True)
	immediate_start = BooleanProperty(False)
	character_name = ObjectProperty()

	@logwrap
	def on_selection(self, *_):
		Logger.debug("App: {} selected".format(self.selection))

	@logwrap
	def on_selected_proxy(self, *_):
		if hasattr(self.selected_proxy, "name"):
			self.selected_proxy_name = str(self.selected_proxy.name)
			return
		selected_proxy = self.selected_proxy
		assert hasattr(selected_proxy, "origin"), "{} has no origin".format(
			type(selected_proxy)
		)
		assert hasattr(selected_proxy, "destination"), (
			"{} has no destination".format(type(selected_proxy))
		)
		origin = selected_proxy.origin
		destination = selected_proxy.destination
		self.selected_proxy_name = (
			str(origin.name) + "->" + str(destination.name)
		)

	def on_character_name(self, name, *_):
		if (
			hasattr(self, "engine")
			and name in self.engine.character
			and (not self.character or self.character.name != name)
		):
			self.character = self.engine.character[name]

	@logwrap
	def _pull_time(self, *_):
		if not hasattr(self, "engine"):
			Clock.schedule_once(self._pull_time, 0)
			return
		branch, turn, tick = self.engine._btt()
		self.branch = branch
		self.turn = turn
		self.tick = tick

	pull_time = trigger(_pull_time)

	@logwrap
	def _really_time_travel(self, branch, turn, tick):
		try:
			self.engine._set_btt(
				branch, turn, tick, cb=self._update_from_time_travel
			)
		except OutOfTimelineError as ex:
			Logger.warning(
				f"App: couldn't time travel to {(branch, turn, tick)}: "
				+ ex.args[0],
				exc_info=ex,
			)
			(self.branch, self.turn, self.tick) = (
				ex.branch_from,
				ex.turn_from,
				ex.tick_from,
			)
		finally:
			self.edit_locked = False
			del self._time_travel_thread

	@logwrap
	def time_travel(self, branch, turn, tick=None):
		if hasattr(self, "_time_travel_thread"):
			return
		self.edit_locked = True
		self._time_travel_thread = Thread(
			target=self._really_time_travel, args=(branch, turn, tick)
		)
		self._time_travel_thread.start()

	@logwrap
	def _really_time_travel_to_tick(self, tick):
		try:
			self.engine._set_btt(
				self.branch, self.turn, tick, cb=self._update_from_time_travel
			)
		except OutOfTimelineError as ex:
			Logger.warning(
				f"App: couldn't time travel to {(self.branch, self.turn, tick)}: "
				+ ex.args[0],
				exc_info=ex,
			)
			(self.branch, self.turn, self.tick) = (
				ex.branch_from,
				ex.turn_from,
				ex.tick_from,
			)
		finally:
			self.edit_locked = False
			if hasattr(self, "_time_travel_thread"):
				del self._time_travel_thread

	@logwrap
	def time_travel_to_tick(self, tick):
		self._time_travel_thread = Thread(
			target=self._really_time_travel_to_tick, args=(tick,)
		)
		self._time_travel_thread.start()

	@logwrap
	def _update_from_time_travel(
		self, command, branch, turn, tick, result, **kwargs
	):
		(self.branch, self.turn, self.tick) = (branch, turn, tick)
		self.mainscreen.update_from_time_travel(
			command, branch, turn, tick, result, **kwargs
		)

	@logwrap
	def set_tick(self, t):
		"""Set my tick to the given value, cast to an integer."""
		self.tick = int(t)

	@logwrap
	def set_turn(self, t):
		"""Set the turn to the given value, cast to an integer"""
		self.turn = int(t)

	@logwrap
	def select_character(self, char: CharacterProxy):
		"""Change my ``character`` to the selected character object if they
		aren't the same.

		"""
		if char == self.character:
			return
		if char.name not in self.mainscreen.graphboards:
			load_kv("elide.graph.board")
			self.mainscreen.graphboards[char.name] = GraphBoard(character=char)
		if char.name not in self.mainscreen.gridboards:
			load_kv("elide.grid.board")
			self.mainscreen.gridboards[char.name] = GridBoard(character=char)
		self.character = char
		self.selected_proxy = self._get_selected_proxy()
		self.engine.eternal["boardchar"] = char.name

	@logwrap
	def build_config(self, config):
		"""Set config defaults"""
		Logger.debug("ElideApp: build_config")
		for sec in "lisien", "elide":
			config.adddefaultsection(sec)
		config.setdefaults(
			"lisien",
			{
				"language": "eng",
				"logfile": "lisien.log",
				"loglevel": "debug",
				"replayfile": "",
				"connect_str": "",
			},
		)
		config.setdefaults(
			"elide",
			{
				"debugger": "no",
				"inspector": "no",
				"user_kv": "yes",
				"play_speed": "1",
				"thing_graphics": json.dumps(
					[
						("Kenney: 1 bit", "kenney1bit.atlas"),
						("RLTiles: Body", "base.atlas"),
						("RLTiles: Basic clothes", "body.atlas"),
						("RLTiles: Armwear", "arm.atlas"),
						("RLTiles: Legwear", "leg.atlas"),
						("RLTiles: Right hand", "hand1.atlas"),
						("RLTiles: Left hand", "hand2.atlas"),
						("RLTiles: Boots", "boot.atlas"),
						("RLTiles: Hair", "hair.atlas"),
						("RLTiles: Beard", "beard.atlas"),
						("RLTiles: Headwear", "head.atlas"),
					]
				),
				"place_graphics": json.dumps(
					[
						("Kenney: 1 bit", "kenney1bit.atlas"),
						("RLTiles: Dungeon", "dungeon.atlas"),
						("RLTiles: Floor", "floor.atlas"),
					]
				),
			},
		)

	def build(self):
		Logger.debug("ElideApp: build")
		self.icon = "icon_24px.png"
		config = self.config

		if config["elide"]["debugger"] == "yes":
			import pdb

			pdb.set_trace()

		self.manager = ScreenManager(transition=NoTransition())
		if config["elide"]["inspector"] == "yes":
			from kivy.core.window import Window
			from kivy.modules import inspector

			inspector.create_inspector(Window, self.manager)
		self.mainmenu = elide.menu.MainMenuScreen(toggle=self._toggler("main"))
		self.manager.add_widget(self.mainmenu)
		if self.immediate_start:
			self.start_game()
		else:
			Clock.schedule_once(self.update_root_viewport, 3)
		return self.manager

	def update_root_viewport(self, *_):
		if not self.root_window:
			Clock.schedule_once(self.update_root_viewport, 0)
			return
		self.root_window.update_viewport()
		Logger.debug("ElideApp: updated root viewport")

	def _pull_lang(self, *_, **kwargs):
		self.strings.language = kwargs["language"]

	def _pull_chars(self, *_, **__):
		self.chars.names = list(self.engine.character)

	def _pull_time_from_signal(self, *_, then, now):
		self.branch, self.turn, self.tick = now
		self.mainscreen.ids.turnscroll.value = self.turn

	def start_subprocess(self, path=None, *_):
		"""Start the lisien core and get a proxy to it

		Must be called before ``init_board``

		"""
		Logger.debug(f"ElideApp: start_subprocess(path={path!r})")
		if hasattr(self, "procman") and hasattr(self.procman, "engine_proxy"):
			raise ChildProcessError("Subprocess already running")
		config = self.config
		enkw = {
			"do_game_start": getattr(self, "do_game_start", False),
		}
		if s := (
			config["lisien"].get("connect_string") or self.connect_string
		):
			s = str(s)
			if "{prefix}" in s:
				enkw["connect_string"] = s.format(prefix=path)
			else:
				Logger.warning("{prefix} not found in " + s)
				enkw["connect_string"] = s
		elif os.path.isfile(os.path.join(path, "world.sqlite3")):
			enkw["connect_string"] = "sqlite:///" + str(
				os.path.join(path, "world.sqlite3")
			)
		workers = config["lisien"].get("workers", "")
		if workers:
			enkw["workers"] = workers
		if config["lisien"].get("logfile"):
			enkw["logfile"] = config["lisien"]["logfile"]
		if config["lisien"].get("loglevel"):
			enkw["loglevel"] = config["lisien"]["loglevel"]
		if config["lisien"].get("replayfile"):
			self._replayfile = open(config["lisien"].get("replayfile"), "at")
			enkw["replay_file"] = self._replayfile
		if self.workers is not None:
			enkw["workers"] = int(self.workers)
		elif workers := config["lisien"].get("workers"):
			enkw["workers"] = int(workers)
		if path:
			os.makedirs(path, exist_ok=True)
		Logger.debug(f"About to start EngineProcessManager with kwargs={enkw}")
		self.procman = EngineProcessManager(
			use_thread=self.use_thread,
		)
		self.engine = engine = self.procman.start(path, **enkw)
		Logger.debug("Got EngineProxy")
		if "boardchar" in engine.eternal:
			self.character_name = engine.eternal["boardchar"]
		elif self.character_name is not None:
			if self.character_name in engine.character:
				self.character = engine.character[self.character_name]
			else:
				self.character = engine.new_character(self.character_name)
		Logger.debug("Pulled character")
		self.pull_time()
		Logger.debug("Pulled time")

		self.engine.time.connect(self._pull_time_from_signal, weak=False)
		self.engine.character.connect(self._pull_chars, weak=False)

		self.strings.store = self.engine.string
		Logger.debug("EngineProxy is ready")
		return engine

	trigger_start_subprocess = trigger(start_subprocess)

	def init_board(self, *_):
		"""Get the board widgets initialized to display the game state

		Must be called after start_subprocess

		"""
		Logger.debug("ElideApp: init_board")
		self.chars.names = char_names = list(self.engine.character)
		for name in char_names:
			if name not in self.mainscreen.graphboards:
				load_kv("elide.graph.board")
				self.mainscreen.graphboards[name] = GraphBoard(
					character=self.engine.character[name]
				)
			if name not in self.mainscreen.gridboards:
				load_kv("elide.grid.board")
				self.mainscreen.gridboards[name] = GridBoard(
					character=self.engine.character[name]
				)
		if "boardchar" in self.engine.eternal:
			self.select_character(
				self.engine.character[self.engine.eternal["boardchar"]]
			)

	def _toggler(self, screenname):
		def tog(*_):
			if self.manager.current == screenname:
				self.manager.current = "mainscreen"
			else:
				self.manager.current = screenname

		return tog

	def _add_screens(self, *_):
		Logger.debug("ElideApp: _add_screens")
		toggler = self._toggler
		config = self.config

		pawndata = json.loads(config["elide"]["thing_graphics"])
		custom_pawns = resource_find("custom_pawn_imgs/custom.atlas")
		if custom_pawns:
			pawndata = [
				["Custom pawns", "custom_pawn_imgs/custom.atlas"]
			] + pawndata

		load_kv("elide.spritebuilder")
		self.pawncfg = elide.spritebuilder.PawnConfigScreen(
			toggle=toggler("pawncfg"),
			data=pawndata,
		)

		spotdata = json.loads(config["elide"]["place_graphics"])
		custom_spots = resource_find("custom_spot_imgs/custom.atlas")
		if custom_spots:
			spotdata = [
				["Custom spots", "custom_spot_imgs/custom.atlas"]
			] + spotdata
		self.spotcfg = elide.spritebuilder.SpotConfigScreen(
			toggle=toggler("spotcfg"),
			data=spotdata,
		)

		load_kv("elide.statcfg")
		self.statcfg = elide.statcfg.StatScreen(toggle=toggler("statcfg"))
		load_kv("elide.rulesview")
		self.rules = elide.rulesview.RulesScreen(toggle=toggler("rules"))

		self.charrules = elide.rulesview.CharacterRulesScreen(
			character=self.character, toggle=toggler("charrules")
		)
		self.bind(character=self.charrules.setter("character"))
		load_kv("elide.charsview")
		self.chars = elide.charsview.CharactersScreen(
			toggle=toggler("chars"), new_board=self.new_board
		)
		self.bind(character_name=self.chars.setter("character_name"))

		def chars_push_character_name(*_):
			self.unbind(character_name=self.chars.setter("character_name"))
			self.character_name = self.chars.character_name
			self.bind(character_name=self.chars.setter("character_name"))

		self.chars.push_character_name = chars_push_character_name
		load_kv("elide.stores")
		self.strings = elide.stores.StringsEdScreen(toggle=toggler("strings"))

		self.funcs = elide.stores.FuncsEdScreen(
			name="funcs", toggle=toggler("funcs")
		)

		self.bind(selected_proxy=self.statcfg.setter("proxy"))
		load_kv("elide.timestream")
		self.timestream = elide.timestream.TimestreamScreen(
			name="timestream", toggle=toggler("timestream")
		)
		load_kv("elide.screen")
		load_kv("elide.charmenu")
		load_kv("elide.statcfg")
		load_kv("elide.stepper")
		load_kv("elide.dialog")
		self.mainscreen = elide.screen.MainScreen(
			use_kv=config["elide"]["user_kv"] == "yes",
			play_speed=int(config["elide"]["play_speed"]),
		)
		if self.mainscreen.statlist:
			self.statcfg.statlist = self.mainscreen.statlist
		self.mainscreen.bind(statlist=self.statcfg.setter("statlist"))
		self.bind(
			selection=self.refresh_selected_proxy,
			character=self.refresh_selected_proxy,
		)
		for wid in (
			self.mainscreen,
			self.pawncfg,
			self.spotcfg,
			self.statcfg,
			self.rules,
			self.charrules,
			self.chars,
			self.strings,
			self.funcs,
			self.timestream,
		):
			self.manager.add_widget(wid)

	def _remove_screens(self):
		Logger.debug("ElideApp: _remove_screens")
		if hasattr(self, "mainscreen"):
			Clock.unschedule(self.mainscreen.play)
		if not hasattr(self, "manager"):
			return
		for widname in (
			"mainscreen",
			"pawncfg",
			"spotcfg",
			"statcfg",
			"rules",
			"charrules",
			"chars",
			"strings",
			"funcs",
			"timestream",
		):
			if not hasattr(self, widname):
				continue
			wid = getattr(self, widname)
			if not isinstance(wid, Screen):
				Logger.info(
					f"ElideApp: not removing {widname} because it's just a mock"
				)
				continue
			self.manager.remove_widget(wid)

	def start_game(self, *_, name=None, cb=None):
		Logger.debug(f"ElideApp: start_game(name={name!r}, cb={cb!r})")
		if hasattr(self, "engine"):
			Logger.error("Already started the game")
			raise RuntimeError("Already started the game")
		game_name = name or self.game_name
		if self.game_name != game_name:
			self.game_name = game_name
		gamedir = os.path.join(self.prefix, self.games_dir, game_name)
		self._add_screens()
		self.manager.current = "mainscreen"
		engine = self.engine = self.start_subprocess(gamedir)
		self.mainscreen.populate()
		self.init_board()
		if cb:
			cb()
		return engine

	def close_game(self, *_, cb=None):
		Logger.debug(f"ElideApp: close_game(cb={cb!r})")
		self.mainmenu.invalidate_popovers()
		if hasattr(self, "manager") and "main" in self.manager.screen_names:
			self.manager.current = "main"
		if hasattr(self, "procman"):
			self.procman.shutdown()
		if hasattr(self, "engine"):
			del self.engine
		else:
			Logger.debug("ElideApp: already closed")
			return  # already closed

		self._copy_log_files()
		game_wd = os.path.join(self.prefix, self.game_name)
		pycache = os.path.join(self.prefix, self.game_name, "__pycache__")
		if os.path.exists(pycache):
			shutil.rmtree(pycache)
		archived = shutil.make_archive(
			self.game_name, "zip", str(game_wd), logger=Logger
		)
		archived_base = os.path.basename(archived)
		os.makedirs(self.games_dir, exist_ok=True)
		if os.path.exists(os.path.join(self.games_dir, archived_base)):
			os.remove(os.path.join(self.games_dir, archived_base))
		os.rename(archived, os.path.join(self.games_dir, archived_base))
		if not hasattr(self, "leave_game"):
			shutil.rmtree(game_wd)
		self._remove_screens()
		if cb:
			cb()

	def update_calendar(self, calendar, past_turns=1, future_turns=5):
		"""Fill in a calendar widget with actual simulation data"""
		Logger.debug(
			f"ElideApp: update_calendar({calendar!r}, "
			f"past_turns={past_turns!r}, "
			f"future_turns={future_turns!r})"
		)
		startturn = self.turn - past_turns
		endturn = self.turn + future_turns
		stats = [
			stat
			for stat in self.selected_proxy
			if isinstance(stat, str)
			and not stat.startswith("_")
			and stat not in ("character", "name", "units", "wallpaper")
		]
		if "_config" in self.selected_proxy:
			stats.append("_config")
		if isinstance(self.selected_proxy, CharStatProxy):
			sched_entity = self.engine.character[self.selected_proxy.name]
		else:
			sched_entity = self.selected_proxy
		calendar.entity = sched_entity
		if startturn == endturn == self.turn:
			# It's the "calendar" that's actually just the current stats
			# of the selected entity, on the left side of elide
			schedule = {stat: [self.selected_proxy[stat]] for stat in stats}
		else:
			schedule = (
				self.engine.handle(
					"get_schedule",
					entity=sched_entity,
					stats=stats,
					beginning=startturn,
					end=endturn,
				),
			)
		calendar.from_schedule(schedule, start_turn=startturn)

	def _set_language(self, lang):
		self.engine.string.language = lang

	def _get_selected_proxy(self):
		Logger.debug("ElideApp: _get_selected_proxy")
		if self.selection is None:
			return self.character.stat
		elif hasattr(self.selection, "proxy"):
			return self.selection.proxy
		elif hasattr(self.selection, "origin") and hasattr(
			self.selection, "destination"
		):
			return self.character.portal[self.selection.origin.name][
				self.selection.destination.name
			]
		else:
			raise ValueError("Invalid selection: {}".format(self.selection))

	def refresh_selected_proxy(self, *_):
		self.selected_proxy = self._get_selected_proxy()

	def on_character(self, *_):
		if not hasattr(self, "mainscreen"):
			Logger.debug("ElideApp: got character before mainscreen")
			Clock.schedule_once(self.on_character, 0)
			return
		if (
			self.character.name not in self.mainscreen.graphboards
			or self.character.name not in self.mainscreen.gridboards
		):
			Logger.debug("ElideApp: got character before boards made")
			Clock.schedule_once(self.on_character, 0)
			return
		Logger.debug("ElideApp: changed character, deselecting")
		if self.character_name != self.character.name:
			self.character_name = self.character.name
		if hasattr(self, "_oldchar"):
			self.mainscreen.graphboards[self._oldchar.name].unbind(
				selection=self.setter("selection")
			)
			self.mainscreen.gridboards[self._oldchar.name].unbind(
				selection=self.setter("selection")
			)
		self.selection = None
		self.mainscreen.graphboards[self.character.name].bind(
			selection=self.setter("selection")
		)
		self.mainscreen.gridboards[self.character.name].bind(
			selection=self.setter("selection")
		)

	def copy_to_shared_storage(
		self,
		filename: str,
		mimetype: str | None = None,
	) -> None:
		Logger.debug(
			f"ElideApp: copy_to_shared_storage({filename!r}, {mimetype!r})"
		)
		try:
			from androidstorage4kivy import SharedStorage
		except ModuleNotFoundError:
			# "shared storage" is just the working directory
			try:
				shutil.copy(
					filename,
					os.path.join(os.path.curdir, os.path.basename(filename)),
				)
			except shutil.SameFileError:
				pass
			return
		if not hasattr(self, "_ss"):
			self._ss = SharedStorage()
		storage = self._ss
		storage.copy_to_shared(filename)

	def _copy_log_files(self):
		Logger.debug("ElideApp: _copy_log_files")
		try:
			from android.storage import app_storage_path
			from jnius import JavaException

			log_dirs = {
				os.path.join(app_storage_path(), "app", ".kivy", "logs")
			}
		except ModuleNotFoundError:
			log_dirs = set()
			JavaException = Exception
		for handler in Logger.handlers:
			if hasattr(handler, "log_dir"):
				log_dir = handler.log_dir
			elif hasattr(handler, "filename"):
				log_dir = os.path.dirname(handler.filename)
			elif hasattr(handler, "baseFilename"):
				log_dir = handler.baseFilename
				if not os.path.isdir(log_dir):
					log_dir = os.path.dirname(log_dir)
			else:
				Logger.error(
					f"ElideApp: handler {handler} (of type {type(handler)})"
					f"has neither log_dir nor filename nor baseFilename"
				)
				continue
			log_dirs.add(log_dir)
		os.makedirs(
			os.path.join(self.prefix, self.game_name, "logs"),
			exist_ok=True,
		)
		failed = []
		for log_dir in log_dirs:
			for logfile in os.listdir(log_dir):
				if logfile.endswith(".log") or (
					logfile.startswith("kivy_") and logfile.endswith(".txt")
				):
					shutil.copy(
						os.path.join(log_dir, logfile),
						str(
							os.path.join(
								self.prefix,
								self.game_name,
								"logs",
								logfile,
							)
						),
					)
					try:
						self.copy_to_shared_storage(
							os.path.join(log_dir, logfile),
							mimetype="text/plain",
						)
					except JavaException:
						failed.append(os.path.join(log_dir, logfile))
			if failed:
				Logger.error("Failed to copy log files: " + ", ".join(failed))
			else:
				Logger.info(
					f"ElideApp: copied log files from {log_dir}"
					f" to {os.path.join(self.prefix, self.game_name, 'logs')} and shared storage"
				)

	@triggered()
	def copy_log_files(self):
		self._copy_log_files()

	def on_pause(self):
		"""Sync the database with the current state of the game."""
		Logger.debug("ElideApp: pausing")
		if hasattr(self, "engine"):
			self.engine.commit()
			Logger.debug("ElideApp: committed")
		if hasattr(self, "strings"):
			self.strings.save()
			Logger.debug("ElideApp: saved strings")
		if hasattr(self, "funcs"):
			self.funcs.save()
			Logger.debug("ElideApp: saved funcs")
		self._copy_log_files()
		Logger.debug("ElideApp: paused")
		return True

	def on_resume(self):
		Logger.debug("ElideApp: resuming")
		self.update_root_viewport()
		return True

	def on_stop(self, *largs):
		"""Sync the database, wrap up the game, and halt."""
		Logger.debug("ElideApp: stopping")
		if hasattr(self, "stopped"):
			return
		self.funcs.save()
		if hasattr(self, "engine"):
			self.close_game()
		self.stopped = True
		Logger.debug("ElideApp: stopped")
		return True

	def delete_selection(self):
		"""Delete both the selected widget and whatever it represents."""
		Logger.debug("ElideApp: delete_selection")
		selection = self.selection
		if selection is None:
			return
		if isinstance(selection, GraphArrow):
			self.mainscreen.boardview.board.rm_arrow(
				selection.origin.name, selection.destination.name
			)
			selection.character.portal[selection.origin.name][
				selection.destination.name
			].delete()
		elif isinstance(selection.proxy, PlaceProxy):
			charn = selection.board.character.name
			self.mainscreen.graphboards[charn].rm_spot(selection.name)
			gridb = self.mainscreen.gridboards[charn]
			if selection.name in gridb.spot:
				gridb.rm_spot(selection.name)
			selection.proxy.delete()
		else:
			assert isinstance(selection.proxy, ThingProxy)
			charn = selection.board.character.name
			self.mainscreen.graphboards[charn].rm_pawn(selection.name)
			self.mainscreen.gridboards[charn].rm_pawn(selection.name)
			selection.proxy.delete()
		self.selection = None
		Logger.debug("ElideApp: selection deleted")

	def new_board(self, name):
		"""Make a graph for a character name, and switch to it."""
		Logger.debug(f"ElideApp: new_board({name!r})")
		char = self.engine.character[name]
		self.mainscreen.graphboards[name] = GraphBoard(character=char)
		self.mainscreen.gridboards[name] = GridBoard(character=char)
		self.character = char
		Logger.debug("ElideApp: made new board for %s", name)

	def on_edit_locked(self, *_):
		Logger.debug(
			"ELiDEApp: "
			+ ("edit locked" if self.edit_locked else "edit unlocked")
		)



# This file is part of Elide, frontend to Lisien, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from collections import defaultdict
from threading import Thread

from kivy.app import App
from kivy.clock import triggered
from kivy.graphics import Color, Line
from kivy.logger import Logger
from kivy.properties import (
	BooleanProperty,
	NumericProperty,
	ObjectProperty,
	StringProperty,
)
from kivy.uix.button import Button
from kivy.uix.recycleview import RecycleView
from kivy.uix.screenmanager import Screen
from kivy.uix.widget import Widget

from .util import logwrap, store_kv


def trigger(func):
	return triggered()(func)


class ThornyRectangle(Button):
	left_margin = NumericProperty(10)
	right_margin = NumericProperty(10)
	top_margin = NumericProperty(10)
	bottom_margin = NumericProperty(10)

	draw_left = BooleanProperty(False)
	draw_right = BooleanProperty(False)
	draw_up = BooleanProperty(False)
	draw_down = BooleanProperty(False)

	branch = StringProperty()
	turn = NumericProperty()

	def __init__(self, **kwargs):
		super().__init__(**kwargs)
		self.bind(
			pos=self._trigger_redraw,
			size=self._trigger_redraw,
			left_margin=self._trigger_redraw,
			right_margin=self._trigger_redraw,
			top_margin=self._trigger_redraw,
			bottom_margin=self._trigger_redraw,
			draw_left=self._trigger_redraw,
			draw_right=self._trigger_redraw,
			draw_up=self._trigger_redraw,
			draw_down=self._trigger_redraw,
		)
		self._trigger_redraw()

	@logwrap(section="ThornyRectangle")
	def collide_point(self, x, y):
		return (
			self.x + self.left_margin < x < self.right - self.right_margin
			and self.y + self.bottom_margin < y < self.top - self.top_margin
		)

	@logwrap(section="ThornyRectangle")
	def _redraw_line(self, enabled, name, point_lambda):
		if enabled:
			points = point_lambda()
			if hasattr(self, name):
				the_line = getattr(self, name)
				the_line.points = points
			else:
				the_line = Line(points=points)
			if the_line not in self.canvas.children:
				self.canvas.add(the_line)
			setattr(self, name, the_line)
		elif hasattr(self, name):
			self.canvas.remove(getattr(self, name))
			delattr(self, name)

	def _get_left_line_points(self):
		return [
			self.x,
			self.center_y,
			self.x + self.left_margin,
			self.center_y,
		]

	def _get_right_line_points(self):
		return [
			self.right - self.right_margin,
			self.center_y,
			self.right,
			self.center_y,
		]

	def _get_top_line_points(self):
		return [
			self.center_x,
			self.top,
			self.center_x,
			self.top - self.top_margin,
		]

	def _get_bottom_line_points(self):
		return [
			self.center_x,
			self.y,
			self.center_x,
			self.y + self.bottom_margin,
		]

	@logwrap(section="ThornyRectangle")
	def _redraw(self, *_):
		self._color = Color(rgba=[1, 1, 1, 1])
		if self._color not in self.canvas.children:
			self.canvas.add(self._color)
		rectpoints = [
			self.x + self.left_margin,
			self.y + self.bottom_margin,
			self.right - self.right_margin,
			self.y + self.bottom_margin,
			self.right - self.right_margin,
			self.top - self.top_margin,
			self.x + self.left_margin,
			self.top - self.top_margin,
			self.x + self.left_margin,
			self.y + self.bottom_margin,
		]
		if hasattr(self, "_rect"):
			self._rect.points = rectpoints
		else:
			self._rect = Line(points=rectpoints)
			self.canvas.add(self._rect)
		self._redraw_line(
			self.draw_left, "_left_line", self._get_left_line_points
		)
		self._redraw_line(
			self.draw_right, "_right_line", self._get_right_line_points
		)
		self._redraw_line(self.draw_up, "_top_line", self._get_top_line_points)
		self._redraw_line(
			self.draw_down, "_bot_line", self._get_bottom_line_points
		)
		self.canvas.ask_update()

	_trigger_redraw = trigger(_redraw)

	@logwrap(section="ThornyRectangle")
	def on_release(self):
		if self.branch is None or self.turn is None:
			return
		app = App.get_running_app()
		app.mainscreen.toggle_timestream()
		self._push_time()

	@triggered(timeout=0.1)
	def _push_time(self, *_):
		app = App.get_running_app()
		app.branch, app.turn = self.branch, self.turn


class Cross(Widget):
	draw_left = BooleanProperty(True)
	draw_right = BooleanProperty(True)
	draw_up = BooleanProperty(True)
	draw_down = BooleanProperty(True)

	def __init__(self, **kwargs):
		super().__init__(**kwargs)
		self.bind(
			draw_left=self._trigger_redraw,
			draw_right=self._trigger_redraw,
			draw_up=self._trigger_redraw,
			draw_down=self._trigger_redraw,
			size=self._trigger_redraw,
			pos=self._trigger_redraw,
		)

	@logwrap(section="Cross")
	def _draw_line(self, enabled, name, get_points):
		if enabled:
			points = get_points()
			if hasattr(self, name):
				getattr(self, name).points = points
			else:
				the_line = Line(points=points)
				setattr(self, name, the_line)
				self.canvas.add(the_line)
		elif hasattr(self, name):
			the_line = getattr(self, name)
			if the_line in self.canvas.children:
				self.canvas.remove(the_line)
			delattr(self, name)

	def _get_left_points(self):
		return [self.x, self.center_y, self.center_x, self.center_y]

	def _get_right_points(self):
		return [self.center_x, self.center_y, self.right, self.center_y]

	def _get_up_points(self):
		return [self.center_x, self.center_y, self.center_x, self.top]

	def _get_down_points(self):
		return [self.center_x, self.center_y, self.center_x, self.y]

	@logwrap(section="Cross")
	def _redraw(self, *_):
		self._draw_line(self.draw_left, "_left_line", self._get_left_points)
		self._draw_line(self.draw_right, "_right_line", self._get_right_points)
		self._draw_line(self.draw_up, "_up_line", self._get_up_points)
		self._draw_line(self.draw_down, "_down_line", self._get_down_points)
		self.canvas.ask_update()

	_trigger_redraw = trigger(_redraw)


class Timestream(RecycleView):
	cols = NumericProperty(
		1
	)  # should be equal to the number of turns on which branches were created + 1


@logwrap(section="elide.timestream")
def _data_and_cols_from_branches(branches):
	start_turn_branches = defaultdict(set)
	end_turn_branches = defaultdict(set)
	branch_split_turns_todo = defaultdict(set)
	branch_split_turns_done = defaultdict(set)
	for branch, (
		parent,
		parent_turn,
		parent_tick,
		end_turn,
		end_tick,
	) in branches.items():
		start_turn_branches[parent_turn].add(branch)
		end_turn_branches[end_turn].add(branch)
		branch_split_turns_todo[parent].add(parent_turn)
	branch_split_turns_todo["trunk"].add(0)
	col2turn = sorted(start_turn_branches.keys() | end_turn_branches.keys())
	if not col2turn:
		return [], 1
	Logger.debug("Timestream: read branch lineage, processing...")
	data = []
	trunk_lineage = branches.pop("trunk")
	sorted_branches = [("trunk", (None, 0, 0, 0, 0))] + sorted(
		branches.items(), key=lambda x: x[1][1]
	)
	branches["trunk"] = trunk_lineage
	for row, (branch, _) in enumerate(sorted_branches):
		for turn in col2turn:
			if branch == "trunk" and turn == 0:
				data.append(
					{
						"widget": "ThornyRectangle",
						"branch": "trunk",
						"turn": 0,
						"draw_left": False,
						"draw_up": False,
						"draw_down": len(start_turn_branches[turn]) > 1,
						"draw_right": bool(branch_split_turns_todo[branch]),
					}
				)
			elif branch in start_turn_branches[turn]:
				here_branches = [
					(branches[b][1], b) for b in start_turn_branches[turn]
				]
				if branch == min(here_branches)[1]:
					data.append(
						{
							"widget": "ThornyRectangle",
							"branch": branch,
							"turn": turn,
							"draw_left": False,
							"draw_up": turn == branches[branch][1],
							"draw_down": len(start_turn_branches[turn]) > 1,
							"draw_right": branches[branch][3] > turn,
						}
					)
				elif branch == max(here_branches)[1]:
					data.append(
						{
							"widget": "ThornyRectangle",
							"branch": branch,
							"turn": turn,
							"draw_left": branches[branch][1] > turn,
							"draw_up": False,
							"draw_down": len(start_turn_branches[turn]) > 1,
							"draw_right": False,
						}
					)
				else:
					data.append(
						{
							"widget": "Cross",
							"draw_left": False,
							"draw_up": True,
							"draw_down": len(start_turn_branches[turn]) > 1,
							"draw_right": branches[branch][3] > turn,
						}
					)
			elif branch in end_turn_branches[turn]:
				data.append(
					{
						"widget": "ThornyRectangle",
						"branch": branch,
						"turn": turn,
						"draw_left": True,
						"draw_up": row > 0 and bool(start_turn_branches[turn]),
						"draw_down": bool(start_turn_branches[turn]),
						"draw_right": False,
					}
				)
			elif branches[branch][1] <= turn:
				here_branches = [
					(branches[b][1], b)
					for b in start_turn_branches[turn]
					if b != branch
				]
				data.append(
					{
						"widget": "Cross",
						"draw_left": turn < branches[branch][3],
						"draw_right": turn < branches[branch][3],
						"draw_up": branch in branches
						and branches[branch][0] in branches
						and branches[branches[branch][0]][3] >= turn,
						"draw_down": bool(here_branches),
					}
				)
			else:
				data.append({"widget": "Widget"})
			branch_split_turns_todo[branch].discard(turn)
			start_turn_branches[turn].discard(branch)
			if turn in end_turn_branches:
				end_turn_branches[turn].discard(branch)
			branch_split_turns_done[branch].add(turn)
		Logger.debug(f"Timestream: processed branch {branch}")
	return data, len(col2turn)


class TimestreamScreen(Screen):
	toggle = ObjectProperty()
	timestream = ObjectProperty()
	_thread: Thread

	@logwrap(section="TimestreamScreen")
	def on_pre_enter(self, *_):
		self.timestream.disabled = True
		self._thread = Thread(target=self._get_data)
		self._thread.start()

	def _get_data(self, *_):
		Logger.debug("Timestream: getting branches")
		engine = App.get_running_app().engine
		data, cols = _data_and_cols_from_branches(engine.handle("branches"))
		self.timestream.cols = cols
		self.timestream.data = data
		Logger.debug("Timestream: loaded!")
		self.timestream.disabled = False


store_kv(
	__name__,
	r"""
<ThornyRectangle>:
	text: f"{self.branch}\n{int(self.turn)}"
<Timestream>:
	key_viewclass: 'widget'
	effect_cls: 'ScrollEffect'
	RecycleGridLayout:
		cols: root.cols
		default_width: 100
		default_height: 100
		default_size_hint: None, None
		height: self.minimum_height
		width: self.minimum_width
		size_hint: None, None
<TimestreamScreen>:
	name: 'timestream'
	timestream: timestream
	BoxLayout:
		orientation: 'vertical'
		Timestream:
			id: timestream
			size_hint_y: 0.95
		BoxLayout:
			size_hint_y: 0.05
			Button:
				text: 'Cancel'
				on_press: root.toggle()
""",
)

if __name__ == "__main__":
	from kivy.base import runTouchApp

	branches = {
		"trunk": (None, 0, 0, 1, 27),
		"trunk1": ("trunk", 0, 2494, 0, 2494),
		"trunk2": ("trunk1", 0, 2494, 1, 27),
		"trunk3": ("trunk", 1, 27, 1, 31),
	}
	data, cols = _data_and_cols_from_branches(branches)
	timestream = Timestream(cols=cols)
	timestream.data = data
	runTouchApp(timestream)



# This file is part of Elide, frontend to Lisien, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Editors for textual data in the database.

The data is accessed via a "store" -- a mapping onto the table, used
like a dictionary. Each of the widgets defined here,
:class:`StringsEditor` and :class:`FuncsEditor`, displays a list of
buttons with which the user may select one of the keys in the store,
and edit its value in a text box.

"""

import re
import string
from ast import parse
from functools import partial
from textwrap import dedent, indent

from kivy.app import App
from kivy.clock import Clock, triggered
from kivy.logger import Logger
from kivy.properties import (
	AliasProperty,
	BooleanProperty,
	ListProperty,
	NumericProperty,
	ObjectProperty,
	StringProperty,
)
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.recycleview import RecycleView
from kivy.uix.recycleview.views import RecycleDataViewBehavior
from kivy.uix.screenmanager import Screen
from kivy.uix.textinput import TextInput
from kivy.uix.togglebutton import ToggleButton

from .util import logwrap, store_kv


def trigger(func):
	return triggered()(func)


class RecycleToggleButton(ToggleButton, RecycleDataViewBehavior):
	"""Toggle button at some index in a RecycleView"""

	index = NumericProperty()

	@logwrap(section="RecycleToggleButton")
	def on_touch_down(self, touch):
		if self.collide_point(*touch.pos):
			return self.parent.select_with_touch(self.index, touch)

	@logwrap(section="RecycleToggleButton")
	def apply_selection(self, rv, index, is_selected):
		if is_selected and index == self.index:
			self.state = "down"
		else:
			self.state = "normal"


class StoreButton(RecycleToggleButton):
	"""RecycleToggleButton to select something to edit in a Store"""

	store = ObjectProperty()
	"""Either a FunctionStore or a StringStore"""
	name = StringProperty()
	"""Name of this particular item"""
	source = StringProperty()
	"""Text of this item"""
	select = ObjectProperty()
	"""Function that gets called with my ``index`` when I'm selected"""

	@logwrap(section="RecycleToggleButton")
	def on_parent(self, *_):
		if self.name == "+":
			self.state = "down"
			self.select(self.index)

	@logwrap(section="RecycleToggleButton")
	def on_state(self, *_):
		if self.state == "down":
			self.select(self.index)


class StoreList(RecycleView):
	"""Holder for a :class:`kivy.uix.listview.ListView` that shows what's
	in a store, using one of the StoreAdapter classes.

	"""

	store = ObjectProperty()
	"""Either a FunctionStore or a StringStore"""
	selection_name = StringProperty()
	"""The ``name`` of the ``StoreButton`` currently selected"""
	boxl = ObjectProperty()
	"""Instance of ``SelectableRecycleBoxLayout``"""

	def __init__(self, **kwargs):
		self._i2name = {}
		self._name2i = {}
		super().__init__(**kwargs)

	@logwrap(section="RecycleToggleButton")
	def on_store(self, *_):
		self.store.connect(self._trigger_redata)
		self.redata()

	@logwrap(section="RecycleToggleButton")
	def on_boxl(self, *_):
		self.boxl.bind(selected_nodes=self._pull_selection)

	@logwrap(section="RecycleToggleButton")
	def _pull_selection(self, *_):
		if not self.boxl.selected_nodes:
			return
		self.selection_name = self._i2name[self.boxl.selected_nodes[0]]

	@logwrap(section="RecycleToggleButton")
	def munge(self, datum):
		i, name = datum
		self._i2name[i] = name
		self._name2i[name] = i
		return {
			"store": self.store,
			"text": str(name),
			"name": name,
			"select": self.ids.boxl.select_node,
			"index": i,
		}

	@logwrap(section="RecycleToggleButton")
	def redata(self, *_, **kwargs):
		"""Update my ``data`` to match what's in my ``store``"""
		select_name = kwargs.get("select_name")
		if not self.store:
			Clock.schedule_once(self.redata)
			return
		self.data = list(
			map(self.munge, enumerate(sorted(self.store._cache.keys())))
		)
		if select_name:
			self._trigger_select_name(select_name)

	@logwrap(section="RecycleToggleButton")
	def _trigger_redata(self, *args, **kwargs):
		part = partial(self.redata, *args, **kwargs)
		if hasattr(self, "_scheduled_redata"):
			Clock.unschedule(self._scheduled_redata)
		self._scheduled_redata = Clock.schedule_once(part, 0)

	@logwrap(section="RecycleToggleButton")
	def select_name(self, name, *_):
		"""Select an item by its name, highlighting"""
		self.boxl.select_node(self._name2i[name])

	@logwrap(section="RecycleToggleButton")
	def _trigger_select_name(self, name):
		part = partial(self.select_name, name)
		if hasattr(self, "_scheduled_select_name"):
			Clock.unschedule(self._scheduled_select_name)
		self._scheduled_select_name = Clock.schedule_once(part, 0)


class LanguageInput(TextInput):
	"""Widget to enter the language you want to edit"""

	screen = ObjectProperty()
	"""The instance of ``StringsEdScreen`` that I'm in"""

	@logwrap(section="RecycleToggleButton")
	def on_focus(self, instance, value, *largs):
		if not value:
			if self.screen.language != self.text:
				self.screen.language = self.text
			self.text = ""


class StringsEdScreen(Screen):
	"""A screen in which to edit strings to be presented to humans

	Needs a ``toggle`` function to switch back to the main screen;
	a ``language`` identifier; and a ``language_setter`` function to be called
	with that ``language`` when changed.

	"""

	toggle = ObjectProperty()
	"""Function to switch back to the main screen"""
	language = StringProperty("eng")
	"""Code identifying the language we're editing"""
	edbox = ObjectProperty()
	"""Widget containing editors for the current string and its name"""

	@logwrap(section="RecycleToggleButton")
	def on_language(self, *_):
		if self.edbox is None:
			Clock.schedule_once(self.on_language, 0)
			return
		self.edbox.storelist.redata()
		if self.store.language != self.language:
			self.store.language = self.language

	@logwrap(section="RecycleToggleButton")
	def on_store(self, *_):
		self.language = self.store.language
		self.store.language.connect(self._pull_language)

	@logwrap(section="RecycleToggleButton")
	def _pull_language(self, *_, language):
		self.language = language

	@logwrap(section="RecycleToggleButton")
	def save(self, *_):
		if self.edbox is None:
			Clock.schedule_once(self.save, 0)
			return
		self.edbox.save()


class Editor(BoxLayout):
	"""Abstract widget for editing strings or functions"""

	name_wid = ObjectProperty()
	"""Text input widget holding the name of the string being edited"""
	store = ObjectProperty()
	"""Proxy to the ``FunctionStore`` or ``StringStore``"""
	disable_text_input = BooleanProperty(False)
	"""Whether to prevent text entry (not name entry)"""
	deletable = BooleanProperty(True)
	"""Whether to show a delete button"""
	# This next is the trigger on the EdBox, which may redata the StoreList
	_trigger_save = ObjectProperty()
	_trigger_delete = ObjectProperty()

	@logwrap(section="RecycleToggleButton")
	def save(self, *_):
		"""Put text in my store, return True if it changed"""
		if self.name_wid is None or self.store is None:
			Logger.debug(
				"{}: Not saving, missing name_wid or store".format(
					type(self).__name__
				)
			)
			return
		if not (self.name_wid.text or self.name_wid.hint_text):
			Logger.debug("{}: Not saving, no name".format(type(self).__name__))
			return
		if (
			self.name_wid.text
			and self.name_wid.text[0]
			in string.digits + string.whitespace + string.punctuation
		):
			# TODO alert the user to invalid name
			Logger.warning(
				"{}: Not saving, invalid name".format(type(self).__name__)
			)
			return
		if hasattr(self, "_do_parse"):
			try:
				parse(self.source)
			except SyntaxError:
				# TODO alert user to invalid source
				Logger.debug(
					"{}: Not saving, couldn't parse".format(
						type(self).__name__
					)
				)
				return
		do_redata = False
		if self.name_wid.text:
			if (
				self.name_wid.hint_text
				and self.name_wid.hint_text != self.name_wid.text
				and hasattr(self.store, self.name_wid.hint_text)
			):
				delattr(self.store, self.name_wid.hint_text)
				do_redata = True
			if (
				not hasattr(self.store, self.name_wid.text)
				or getattr(self.store, self.name_wid.text) != self.source
			):
				Logger.debug("{}: Saving!".format(type(self).__name__))
				setattr(self.store, self.name_wid.text, self.source)
				do_redata = True
		elif self.name_wid.hint_text:
			if (
				not hasattr(self.store, self.name_wid.hint_text)
				or getattr(self.store, self.name_wid.hint_text) != self.source
			):
				Logger.debug("{}: Saving!".format(type(self).__name__))
				setattr(self.store, self.name_wid.hint_text, self.source)
				do_redata = True
		return do_redata

	@logwrap(section="RecycleToggleButton")
	def delete(self, *_):
		"""Remove the currently selected item from my store"""
		key = self.name_wid.text or self.name_wid.hint_text
		if not hasattr(self.store, key):
			# TODO feedback about missing key
			return
		delattr(self.store, key)
		try:
			return min(kee for kee in dir(self.store) if kee > key)
		except ValueError:
			return "+"


class StringInput(Editor):
	"""Editor for human-readable strings"""

	validate_name_input = ObjectProperty()
	"""Boolean function for checking if a string name is acceptable"""

	@logwrap(section="StringInput")
	def on_name_wid(self, *_):
		if not self.validate_name_input:
			Clock.schedule_once(self.on_name_wid, 0)
			return
		self.name_wid.bind(text=self.validate_name_input)

	@logwrap(section="StringInput")
	def _get_name(self):
		if self.name_wid:
			return self.name_wid.text

	@logwrap(section="StringInput")
	def _set_name(self, v, *_):
		if not self.name_wid:
			Clock.schedule_once(partial(self._set_name, v), 0)
			return
		self.name_wid.text = v

	name = AliasProperty(_get_name, _set_name)

	@logwrap(section="StringInput")
	def _get_source(self):
		if "string" not in self.ids:
			return ""
		return self.ids.string.text

	@logwrap(section="StringInput")
	def _set_source(self, v, *args):
		if "string" not in self.ids:
			Clock.schedule_once(partial(self._set_source, v), 0)
			return
		self.ids.string.text = v

	source = AliasProperty(_get_source, _set_source)


class EdBox(BoxLayout):
	"""Box containing most of an editor's screen

	Has a StoreList and an Editor, which in turn holds a name field and a big text entry box.

	"""

	storelist = ObjectProperty()
	"""An instance of ``StoreList``"""
	editor = ObjectProperty()
	"""An instance of a subclass of ``Editor``"""
	store = ObjectProperty()
	"""Proxy to the store I represent"""
	store_name = StringProperty()
	"""Name of my store, so I can get it from the engine"""
	data = ListProperty()
	"""Dictionaries describing widgets in my ``storelist``"""
	toggle = ObjectProperty()
	"""Function to show or hide my screen"""
	disable_text_input = BooleanProperty(False)
	"""Set to ``True`` to prevent entering text in the editor"""
	selection_name = StringProperty()

	@logwrap(section="EdBox")
	def on_store_name(self, *_):
		app = App.get_running_app()
		if not hasattr(app, "engine"):
			Clock.schedule_once(self.on_store_name, 0)
			return
		self.store = getattr(app.engine, self.store_name)

	@logwrap(section="EdBox")
	def on_storelist(self, *_):
		self.storelist.bind(selection_name=self.setter("selection_name"))

	@trigger
	@logwrap(section="EdBox")
	def validate_name_input(self, *_):
		self.disable_text_input = not (
			self.valid_name(self.editor.name_wid.hint_text)
			or self.valid_name(self.editor.name_wid.text)
		)

	@logwrap(section="EdBox")
	def on_selection_name(self, _, selection_name):
		if selection_name == self.editor.name_wid.hint_text:
			return
		self.save()
		# The + button at the top is for adding an entry yet unnamed, so don't display hint text for it
		self.editor.name_wid.hint_text = selection_name.strip("+")
		self.editor.name_wid.text = ""
		try:
			self.editor.source = getattr(
				self.store, self.editor.name_wid.hint_text
			)
		except AttributeError:
			self.editor.source = self.get_default_text(
				self.editor.name_wid.hint_text
			)
		self.disable_text_input = not self.valid_name(
			self.editor.name_wid.hint_text
		)
		if hasattr(self, "_lock_save"):
			del self._lock_save

	@logwrap(section="EdBox")
	def dismiss(self, *_):
		self.save()
		self.toggle()

	@logwrap(section="EdBox")
	def save(self, *_, name=None):
		if not self.editor or not self.store:
			return
		if hasattr(self, "_lock_save"):
			return
		self._lock_save = True
		save_select = self.editor.save()
		if save_select:
			self.storelist.redata(select_name=name)
		else:
			del self._lock_save

	@logwrap(section="EdBox")
	def _trigger_save(self, name=None):
		part = partial(self.save, name=name)
		Clock.unschedule(part)
		Clock.schedule_once(part, 0)

	@logwrap(section="EdBox")
	def delete(self, *_):
		if not self.editor:
			return
		if hasattr(self, "_lock_save"):
			return
		self._lock_save = True
		del_select = self.editor.delete()
		if del_select:
			self.storelist.redata(del_select)
		else:
			del self._lock_save

	_trigger_delete = trigger(delete)

	def on_store(self, *_):
		pass


class StringNameInput(TextInput):
	"""Small text box for the names of strings"""

	_trigger_save = ObjectProperty()

	@logwrap(section="StringNameInput")
	def on_focus(self, inst, val, *largs):
		if self.text and not val:
			self._trigger_save(self.text)


class StringsEdBox(EdBox):
	"""Box containing most of the strings editing screen

	Contains the storelist and the editor, which in turn contains the string name input
	and a bigger input field for the string itself.

	"""

	language = StringProperty("eng")

	@staticmethod
	def get_default_text(newname):
		return ""

	@staticmethod
	def valid_name(name):
		return name and name[0] != "+"


sig_ex = re.compile(r"^ *def .+?\((.+)\):$")


class FunctionNameInput(TextInput):
	"""Input for the name of a function

	Filters out illegal characters.

	"""

	_trigger_save = ObjectProperty()

	@logwrap(section="FunctionNameInput")
	def insert_text(self, s, from_undo=False):
		if self.text == "":
			if s[0] not in (string.ascii_letters + "_"):
				return
		return super().insert_text(
			"".join(
				c
				for c in s
				if c in (string.ascii_letters + string.digits + "_")
			)
		)

	@logwrap(section="FunctionNameInput")
	def on_focus(self, inst, val, *_):
		if not val:
			self._trigger_save(self.text)


@logwrap(section="elide.stores")
def munge_source(v):
	"""Take Python source code, return a pair of its parameters and the rest of it dedented"""
	lines = v.split("\n")
	if not lines:
		return tuple(), ""
	firstline = lines[0].lstrip()
	while firstline == "" or firstline[0] == "@":
		del lines[0]
		firstline = lines[0].lstrip()
	if not lines:
		return tuple(), ""
	params = tuple(
		parm.strip() for parm in sig_ex.match(lines[0]).group(1).split(",")
	)
	del lines[0]
	if not lines:
		return params, ""
	# hack to allow 'empty' functions
	if lines and lines[-1].strip() == "pass":
		del lines[-1]
	return params, dedent("\n".join(lines))


class FuncEditor(Editor):
	"""The editor widget for working with any particular function.

	Contains a one-line field for the function's name and a multi-line
	field for its code.

	"""

	storelist = ObjectProperty()
	"""Instance of ``StoreList`` that shows all the functions you can edit"""
	codeinput = ObjectProperty()
	params = ListProperty(["obj"])
	name = StringProperty()
	_text = StringProperty()
	_do_parse = True

	def _get_source(self):
		code = self.get_default_text(self.name)
		if self._text:
			code += indent(self._text, " " * 4)
		else:
			code += " " * 4 + "pass"
		return code.rstrip(" \n\t")

	@logwrap(section="FuncEditor")
	def _set_source(self, v):
		if not self.codeinput:
			Clock.schedule_once(partial(self._set_source, v), 0)
			return
		self.codeinput.unbind(text=self.setter("_text"))
		self.params, self.codeinput.text = munge_source(str(v))
		self.codeinput.bind(text=self.setter("_text"))

	source = AliasProperty(_get_source, _set_source, bind=("_text", "params"))

	def get_default_text(self, name):
		if not name or name == "+":
			name = "a"
		return "def {}({}):\n".format(name, ", ".join(self.params))

	def on_codeinput(self, *args):
		self._text = self.codeinput.text
		self.codeinput.bind(text=self.setter("_text"))


class FuncsEdBox(EdBox):
	"""Widget for editing the Python source of funcs to be used in lisien sims.

	Contains a list of functions in the store it's about, next to a
	FuncEditor showing the source of the selected one, and a close button.

	"""

	def get_default_text(self, newname):
		return self.editor.get_default_text(newname)

	@staticmethod
	def valid_name(name):
		return (
			name
			and name[0]
			not in string.digits + string.whitespace + string.punctuation
		)

	@logwrap(section="FuncsEdBox")
	def on_data(self, *_):
		app = App.get_running_app()
		if app is None:
			return
		app.rules.rulesview.set_functions(
			self.store_name, map(app.rules.rulesview.inspect_func, self.data)
		)


class FuncsEdScreen(Screen):
	"""Screen containing three FuncsEdBox

	Triggers, prereqs, and actions.

	"""

	toggle = ObjectProperty()

	@logwrap(section="FuncsEdScreen")
	def save(self, *args):
		self.ids.triggers.save()
		self.ids.prereqs.save()
		self.ids.actions.save()


store_kv(
	__name__,
	"""
#: import py3lexer pygments.lexers.Python3Lexer
<StoreButton>:
	size_hint_y: None
	height: 30
<StoreList>:
	viewclass: 'StoreButton'
	boxl: boxl
	SelectableRecycleBoxLayout:
		id: boxl
		default_size: None, dp(56)
		default_size_hint: 1, None
		height: self.minimum_height
		size_hint_y: None
		orientation: 'vertical'
<StringInput>:
	name_wid: stringname
	orientation: 'vertical'
	BoxLayout:
		size_hint_y: 0.07
		Label:
			id: title
			text: 'Title: '
			size_hint_x: None
			width: self.texture_size[0]
		StringNameInput:
			id: stringname
			multiline: False
			write_tab: False
			_trigger_save: root._trigger_save
		Button:
			text: 'del'
			size_hint_x: 0.1
			on_release: root._trigger_delete()
	TextInput:
		id: string
		disabled: root.disable_text_input
<StringsEdBox>:
	editor: strings_ed
	storelist: strings_list
	orientation: 'vertical'
	BoxLayout:
		orientation: 'horizontal'
		StoreList:
			id: strings_list
			size_hint_x: 0.2
			store: root.store
		StringInput:
			id: strings_ed
			store: root.store
			disable_text_input: root.disable_text_input
			validate_name_input: root.validate_name_input
			_trigger_save: root._trigger_save
			_trigger_delete: root._trigger_delete
<StringsEdScreen>:
	name: 'strings'
	edbox: edbox
	BoxLayout:
		orientation: 'vertical'
		StringsEdBox:
			id: edbox
			toggle: root.toggle
			store_name: 'string'
			language: root.language
		BoxLayout:
			size_hint_y: 0.05
			Label:
				text_size: self.size
				halign: 'right'
				valign: 'middle'
				text: 'Language: '
				size_hint_x: 0.2
			LanguageInput:
				id: language
				screen: root
				hint_text: root.language
				write_tab: False
				multiline: False
			Button:
				text: 'Close'
				on_release: edbox.dismiss()
<Py3CodeInput@CodeInput>:
	lexer: py3lexer()
<FuncEditor>:
	codeinput: code
	name_wid: funname
	orientation: 'vertical'
	BoxLayout:
		orientation: 'horizontal'
		size_hint_y: None
		height: funname.height
		Py3CodeInput:
			id: imafunction
			text: 'def'
			disabled: True
			size_hint: (None, None)
			height: self.line_height + self.font_size
			width: self.font_size * len(self.text)
			background_disabled_normal: ''
			disabled_foreground_color: self.foreground_color
		FunctionNameInput:
			id: funname
			size_hint: (0.4, None)
			height: self.line_height + self.font_size
			multiline: False
			write_tab: False
			_trigger_save: root._trigger_save
			on_text: root.validate_name_input(self.text)
			hint_text: root.name
		Py3CodeInput:
			id: params
			text: '({}):'.format(', '.join(root.params))
			disabled: True
			size_hint_y: None
			height: self.line_height + self.font_size
			background_disabled_normal: ''
			disabled_foreground_color: self.foreground_color
		Button:
			text: 'del'
			size_hint_x: 0.1 if not self.disabled else 0.0
			background_disabled_normal: ''
			background_disabled_down: ''
			disabled_color: (0., 0., 0., 0.)
			background_disabled_color: (0., 0., 0., 0.)
			on_release: root._trigger_delete()
			disabled: not root.deletable
	BoxLayout:
		orientation: 'horizontal'
		Label:
			canvas:
				Color:
					rgba: params.background_color
				Rectangle:
					pos: self.pos
					size: self.size
				Color:
					rgba: [1., 1., 1., 1.]
			# PEP8 standard indentation width is 4 spaces
			text: ' ' * 4
			size_hint_x: None
			width: self.texture_size[0]
		Py3CodeInput:
			id: code
			disabled: root.disable_text_input
<FuncsEdBox>:
	editor: funcs_ed
	storelist: funcs_list
	orientation: 'vertical'
	data: [(item['name'], self.store.get_source(item['name'])) for item in funcs_list.data[1:]]
	BoxLayout:
		orientation: 'horizontal'
		BoxLayout:
			orientation: 'vertical'
			size_hint_x: 0.2
			StoreList:
				id: funcs_list
				store: root.store
		FuncEditor:
			id: funcs_ed
			name: root.selection_name
			store: root.store
			storelist: funcs_list
			disable_text_input: root.disable_text_input
			validate_name_input: root.validate_name_input
			_trigger_save: root._trigger_save
			_trigger_delete: root._trigger_delete
	BoxLayout:
		size_hint_y: 0.07
		Button:
			text: 'Close'
			on_release: root.dismiss()
<FuncsEdScreen>:
	name: 'funcs'
	TabbedPanel:
		default_tab: action
		TabbedPanelItem:
			id: trigger
			text: 'Trigger'
			on_state: triggers.save()
			FuncsEdBox:
				id: triggers
				toggle: root.toggle
				store_name: 'trigger'
		TabbedPanelItem:
			id: prereq
			text: 'Prereq'
			on_state: prereqs.save()
			FuncsEdBox:
				id: prereqs
				toggle: root.toggle
				store_name: 'prereq'
		TabbedPanelItem:
			id: action
			text: 'Action'
			on_state: actions.save()
			FuncsEdBox:
				id: actions
				toggle: root.toggle
				store_name: 'action'
""",
)
